/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/ioctl.h>
#include <sys/types.h>
#include <assert.h>
#include <err.h>
#include <stdbool.h>
#include <stdlib.h>

#if defined (__linux__)
#include <linux/fb.h>
#include <linux/kd.h>
#include <linux/vt.h>
#elif defined (__FreeBSD__)
#include <osreldate.h>
#if __FreeBSD_version >= 410000
#include <sys/consio.h>
#else
#include <machine/console.h>
#endif
#elif defined (__NetBSD__) || defined (__OpenBSD__)
#include <time.h>
#include <dev/wscons/wsconsio.h>
#ifndef _DEV_WSCONS_WSDISPLAY_USL_IO_H_
#include <dev/wscons/wsdisplay_usl_io.h>
#define _DEV_WSCONS_WSDISPLAY_USL_IO_H_
#endif
#endif

#include "privilege.h"

static bool initialized;

static void finalize(void);
static void finalizer(void);

void console_initialize(void)
{
	assert(!initialized);

	atexit(finalizer);
	initialized = true;
}

static void finalize(void)
{
	assert(initialized);

	initialized = false;
}

static void finalizer(void)
{
	finalize();
}

int console_getActive(void)
{
#if defined (__linux__)
	struct vt_stat vt_stat;
	int result;

	assert(initialized);

	result = 0;
	if (ioctl(STDIN_FILENO, VT_GETSTATE, &vt_stat) != -1)
		result = vt_stat.v_active;
	return result;
#elif defined (__FreeBSD__) || defined (__NetBSD__) || defined (__OpenBSD__)
	int result, active;

	assert(initialized);

	result = 0;
	if (ioctl(STDIN_FILENO, VT_GETACTIVE, &active) != -1)
		result = active;
	return result;
#else
	#error not implement
#endif
}

int console_getMode(void)
{
	int mode;

	assert(initialized);

#if defined (__linux__) || defined (__FreeBSD__)
	if (ioctl(STDIN_FILENO, KDGETMODE, &mode) == -1)
		errx(1, "Could not get console mode.");
#elif defined (__NetBSD__) || defined (__OpenBSD__)
	if (ioctl(STDIN_FILENO, WSDISPLAYIO_GMODE, &mode) == -1)
		errx(1, "Could not get console mode.");
#else
	#error not implement
#endif
	return mode;
}

void console_setMode(int mode)
{
#if defined (__linux__) || defined (__FreeBSD__)
	assert(initialized);

	if (ioctl(STDIN_FILENO, KDSETMODE, mode) == -1)
		errx(1, "Could not set console mode.");
#elif defined (__NetBSD__) || defined (__OpenBSD__)
	int m;

	assert(initialized);

	m = mode;
	if (ioctl(STDIN_FILENO, WSDISPLAYIO_SMODE, &m) == -1)
		errx(1, "Could not set console mode.");
#else
	#error not implement
#endif
}

