/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM/ME - J framebuffer terminal/Multilingual Enhancement -
 * Copyright (C) 2003 Fumitoshi UKAI <ukai@debian.or.jp>
 * Copyright (C) 1999 Noritoshi MASUICHI <nmasu@ma3.justnet.ne.jp>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <string.h>

#if defined (__linux__)
#include <endian.h>
#elif defined (__FreeBSD__) || defined (__NetBSD__) || defined (__OpenBSD__)
#include <sys/stdint.h> /* OpenBSD */
#include <sys/endian.h>
#endif

#include "fbaccess.h"
#include "palette.h"
#ifdef ENABLE_SPLASH_SCREEN
#include "splash.h"
#endif

#ifndef min
#define min(a, b) (((a) > (b)) ? (b) : (a))
#endif

#ifdef ENABLE_8BPP
/* 8 bpp */
void fbaccess_fillRect_8bpp(TFrameBuffer *p, u_int sx, u_int sy,
			    u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_char *line;

	for (y = sy; y < min(sy + ly, p->height); y++) {
#if 0
		memset(p->smem + y * p->bytePerLine + sx, color, lx);
#endif
		line = p->smem + y * p->bytePerLine + sx;
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
			line[x] = color;
		}
	}
}

void fbaccess_clearAll_8bpp(TFrameBuffer *p, u_char color)
{
	memset(p->smem, color, p->slen);
}

void fbaccess_overlay_8bpp(TFrameBuffer *p, u_int xd, u_int yd,
			   const u_char *ps, u_int lx, u_int ly,
			   u_int gap, u_char color)
{
	u_int y;
	u_char *wp;
	const u_char *tps;
	u_int i;
	u_char sb;

	for (y = yd; y < yd + ly; y++) {
		tps = ps;
		wp = p->smem + y * p->bytePerLine + xd;
		for (i = lx; i >= 8; i -= 8) {
			sb = *tps++;
			if (sb & 0x80) wp[0] = color;
			if (sb & 0x40) wp[1] = color;
			if (sb & 0x20) wp[2] = color;
			if (sb & 0x10) wp[3] = color;
			if (sb & 0x08) wp[4] = color;
			if (sb & 0x04) wp[5] = color;
			if (sb & 0x02) wp[6] = color;
			if (sb & 0x01) wp[7] = color;
			wp += 8;
		}
		if (i != 0) {
			sb = *tps++;
			switch (i) {
			case 7: if (sb & 0x02) wp[6] = color;
				/* FALLTHROUGH */
			case 6: if (sb & 0x04) wp[5] = color;
				/* FALLTHROUGH */
			case 5: if (sb & 0x08) wp[4] = color;
				/* FALLTHROUGH */
			case 4: if (sb & 0x10) wp[3] = color;
				/* FALLTHROUGH */
			case 3: if (sb & 0x20) wp[2] = color;
				/* FALLTHROUGH */
			case 2: if (sb & 0x40) wp[1] = color;
				/* FALLTHROUGH */
			case 1: if (sb & 0x80) wp[0] = color;
				/* FALLTHROUGH */
			default: break;
			}
		}
		ps += gap;
	}
}

void fbaccess_reverse_8bpp(TFrameBuffer *p, u_int sx, u_int sy,
			   u_int lx, u_int ly, u_char color)
{
	u_int y;
	u_int x;

	for (y = sy; y < min(sy + ly, p->height); y++)
		for (x = sx; x < sx + lx; x++)
			p->smem[y * p->bytePerLine + x] ^= color;
}
#endif /* ENABLE_8BPP */


#ifdef ENABLE_15BPP
/* 15 bpp */
void fbaccess_fillRect_15bpp(TFrameBuffer *p, u_int sx, u_int sy,
			     u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_short *d;
	u_short icol;

	icol = palette_getTrueColor16(color);
	for (y = sy; y < min(sy + ly, p->height); y++) {
		d = (u_short *)(p->smem + y * p->bytePerLine + sx * 2);
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
#ifdef ENABLE_SPLASH_SCREEN
			if (gSplash.enable && icol == 0 &&
			    SPLASH_INRANGE(gSplash.width, gSplash.height,
					   sx + x, y))
				*d++ = SPLASH_GET_PIXEL(gSplash, sx + x, y);
			else
				*d++ = icol;
#else
			*d++ = icol;
#endif
		}
	}
}

void fbaccess_clearAll_15bpp(TFrameBuffer *p, u_char color)
{
	u_int lp;
	u_short *d;
	u_short icol;

	d = (u_short *)(p->smem);
	icol = palette_getTrueColor16(color);
	for (lp = 0; lp < ((p->slen) / 2); lp++)
		d[lp] = icol;
}

void fbaccess_overlay_15bpp(TFrameBuffer *p, u_int xd, u_int yd,
			    const u_char *ps, u_int lx, u_int ly,
			    u_int gap, u_char color)
{
	u_int y;
	u_short *wp;
	const u_char *tps;
	u_int i;
	u_char sb;
	u_short icol;

	icol = palette_getTrueColor16(color);
	for (y = yd; y < yd + ly; y++) {
		tps = ps;
		wp = (u_short *)(p->smem + y * p->bytePerLine + xd * 2);
		for (i = lx; i >= 8; i -= 8) {
			sb = *tps++;
			if (sb & 0x80) wp[0] = icol;
			if (sb & 0x40) wp[1] = icol;
			if (sb & 0x20) wp[2] = icol;
			if (sb & 0x10) wp[3] = icol;
			if (sb & 0x08) wp[4] = icol;
			if (sb & 0x04) wp[5] = icol;
			if (sb & 0x02) wp[6] = icol;
			if (sb & 0x01) wp[7] = icol;
			wp += 8;
		}
		if (i != 0) {
			sb = *tps++;
			switch (i) {
			case 7: if (sb & 0x02) wp[6] = icol;
				/* FALLTHROUGH */
			case 6: if (sb & 0x04) wp[5] = icol;
				/* FALLTHROUGH */
			case 5: if (sb & 0x08) wp[4] = icol;
				/* FALLTHROUGH */
			case 4: if (sb & 0x10) wp[3] = icol;
				/* FALLTHROUGH */
			case 3: if (sb & 0x20) wp[2] = icol;
				/* FALLTHROUGH */
			case 2: if (sb & 0x40) wp[1] = icol;
				/* FALLTHROUGH */
			case 1: if (sb & 0x80) wp[0] = icol;
				/* FALLTHROUGH */
			default: break;
			}
		}
		ps += gap;
	}
}

void fbaccess_reverse_15bpp(TFrameBuffer *p, u_int sx, u_int sy,
			    u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_short *d;
	u_short icol;

	icol = palette_getTrueColor16(color);
	for (y = sy; y < min(sy + ly, p->height); y++) {
		d = (u_short *)(p->smem + y * p->bytePerLine + sx * 2);
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
			d[x] ^= icol;
		}
	}
}
#endif /* ENABLE_15BPP */

#ifdef ENABLE_16BPP
/* 16 bpp */
void fbaccess_fillRect_16bpp(TFrameBuffer *p, u_int sx, u_int sy,
			     u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_short *d;
	u_short icol;

	icol = palette_getTrueColor16(color);
	for (y = sy; y < min(sy + ly, p->height); y++) {
		d = (u_short *)(p->smem + y * p->bytePerLine + sx * 2);
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
#ifdef ENABLE_SPLASH_SCREEN
			if (gSplash.enable && icol == 0 &&
			    SPLASH_INRANGE(gSplash.width, gSplash.height,
					   sx + x, y))
				*d++ = SPLASH_GET_PIXEL(gSplash, sx + x, y);
			else
				*d++ = icol;
#else
			*d++ = icol;
#endif
		}
	}
}

void fbaccess_clearAll_16bpp(TFrameBuffer *p, u_char color)
{
	u_int lp;
	u_short *d;
	u_short icol;

	d = (u_short *)(p->smem);
	icol = palette_getTrueColor16(color);
	for (lp = 0; lp < ((p->slen) / 2); lp++)
		d[lp] = icol;
}

void fbaccess_overlay_16bpp(TFrameBuffer *p, u_int xd, u_int yd,
			    const u_char *ps, u_int lx, u_int ly,
			    u_int gap, u_char color)
{
	u_int y;
	u_short *wp;
	const u_char *tps;
	u_int i;
	u_char sb;
	u_short icol;

	icol = palette_getTrueColor16(color);
	for (y = yd; y < yd + ly; y++) {
		tps = ps;
		wp = (u_short *)(p->smem + y * p->bytePerLine + xd * 2);
		for (i = lx; i >= 8; i -= 8) {
			sb = *tps++;
			if (sb & 0x80) wp[0] = icol;
			if (sb & 0x40) wp[1] = icol;
			if (sb & 0x20) wp[2] = icol;
			if (sb & 0x10) wp[3] = icol;
			if (sb & 0x08) wp[4] = icol;
			if (sb & 0x04) wp[5] = icol;
			if (sb & 0x02) wp[6] = icol;
			if (sb & 0x01) wp[7] = icol;
			wp += 8;
		}
		if (i != 0) {
			sb = *tps++;
			switch (i) {
			case 7: if (sb & 0x02) wp[6] = icol;
				/* FALLTHROUGH */
			case 6: if (sb & 0x04) wp[5] = icol;
				/* FALLTHROUGH */
			case 5: if (sb & 0x08) wp[4] = icol;
				/* FALLTHROUGH */
			case 4: if (sb & 0x10) wp[3] = icol;
				/* FALLTHROUGH */
			case 3: if (sb & 0x20) wp[2] = icol;
				/* FALLTHROUGH */
			case 2: if (sb & 0x40) wp[1] = icol;
				/* FALLTHROUGH */
			case 1: if (sb & 0x80) wp[0] = icol;
				/* FALLTHROUGH */
			default: break;
			}
		}
		ps += gap;
	}
}

void fbaccess_reverse_16bpp(TFrameBuffer *p, u_int sx, u_int sy,
			    u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_short *d;
	u_short icol;

	icol = palette_getTrueColor16(color);
	for (y = sy; y < min(sy + ly, p->height); y++) {
		d = (u_short *)(p->smem + y * p->bytePerLine + sx * 2);
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
			d[x] ^= icol;
		}
	}
}
#endif /* ENABLE_16BPP */

#ifdef ENABLE_24BPP
/* 24 bpp */
void fbaccess_fillRect_24bpp(TFrameBuffer *p, u_int sx, u_int sy,
			     u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_char *d;
	u_char t, m, b;
	u_int icol;
	u_char *pfx;

	icol = palette_getTrueColor32(color);
#if __BYTE_ORDER == __BIG_ENDIAN
	b = 0xff & (icol >> 16); m = 0xff & (icol >> 8); t = 0xff & icol;
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	t = 0xff & (icol >> 16); m = 0xff & (icol >> 8); b = 0xff & icol;
#else
	#error FIXME : No endianness ?
#endif

	for (y = sy; y < sy + ly; y++) {
		d = p->smem + y * p->bytePerLine + sx * 3;
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
#ifdef ENABLE_SPLASH_SCREEN
			if (gSplash.enable && icol == 0 &&
			    SPLASH_INRANGE(gSplash.width, gSplash.height,
					   sx + x, y)) {
				pfx = SPLASH_GET_PIXEL_24(gSplash, sx + x, y);
				*d++ = *(pfx + 2);
				*d++ = *(pfx + 1);
				*d++ = *(pfx + 0);
			} else {
				*d++ = t;
				*d++ = m;
				*d++ = b;
			}
#else
			*d++ = t;
			*d++ = m;
			*d++ = b;
#endif
		}
	}
}

void fbaccess_clearAll_24bpp(TFrameBuffer *p, u_char color)
{
	u_int lp;
	u_char t, m, b;
	u_int icol;

	icol = palette_getTrueColor32(color);
#if __BYTE_ORDER == __BIG_ENDIAN
	b = 0xff & (icol >> 16); m = 0xff & (icol >> 8); t = 0xff & icol;
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	t = 0xff & (icol >> 16); m = 0xff & (icol >> 8); b = 0xff & icol;
#else
	#error FIXME : No endianness ?
#endif

	for (lp = 0; lp < ((p->slen) - 2); lp += 3) {
		p->smem[lp] = t;
		p->smem[lp + 1] = m;
		p->smem[lp + 2] = b;
	}
}

void fbaccess_overlay_24bpp(TFrameBuffer *p, u_int xd, u_int yd,
			    const u_char *ps, u_int lx, u_int ly,
			    u_int gap, u_char color)
{
	u_int y, i, sb;
	u_char *wp;
	const u_char *tps;
	u_char t, m, b;
	u_int icol;

	icol = palette_getTrueColor32(color);
#if __BYTE_ORDER == __BIG_ENDIAN
	b = 0xff & (icol >> 16); m = 0xff & (icol >> 8); t = 0xff & icol;
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	t = 0xff & (icol >> 16); m = 0xff & (icol >> 8); b = 0xff & icol;
#else
	#error FIXME : No endianness ?
#endif

	for (y = yd; y < yd + ly; y++) {
		tps = ps;
		wp = p->smem + y * p->bytePerLine + xd * 3;
		for (i = lx; i >= 8; i -= 8) {
			sb = *tps++;
			if (sb & 0x80) wp[0] = t, wp[1] = m, wp[2] = b;
			if (sb & 0x40) wp[3] = t, wp[4] = m, wp[5] = b;
			if (sb & 0x20) wp[6] = t, wp[7] = m, wp[8] = b;
			if (sb & 0x10) wp[9] = t, wp[10] = m, wp[11] = b;
			if (sb & 0x08) wp[12] = t, wp[13] = m, wp[14] = b;
			if (sb & 0x04) wp[15] = t, wp[16] = m, wp[17] = b;
			if (sb & 0x02) wp[18] = t, wp[19] = m, wp[20] = b;
			if (sb & 0x01) wp[21] = t, wp[22] = m, wp[23] = b;
			wp += 24;
		}
		if (i != 0) {
			sb = *tps++;
			switch (i) {
			case 7: if (sb & 0x02) wp[18] = t, wp[19] = m, wp[20] = b;
				/* FALLTHROUGH */
			case 6: if (sb & 0x04) wp[15] = t, wp[16] = m, wp[17] = b;
				/* FALLTHROUGH */
			case 5: if (sb & 0x08) wp[12] = t, wp[13] = m, wp[14] = b;
				/* FALLTHROUGH */
			case 4: if (sb & 0x10) wp[9] = t, wp[10] = m, wp[11] = b;
				/* FALLTHROUGH */
			case 3: if (sb & 0x20) wp[6] = t, wp[7] = m, wp[8] = b;
				/* FALLTHROUGH */
			case 2: if (sb & 0x40) wp[3] = t, wp[4] = m, wp[5] = b;
				/* FALLTHROUGH */
			case 1: if (sb & 0x80) wp[0] = t, wp[1] = m, wp[2] = b;
				/* FALLTHROUGH */
			default: break;
			}
		}
		ps += gap;
	}
}

void fbaccess_reverse_24bpp(TFrameBuffer *p, u_int sx, u_int sy,
			    u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_char t, m, b;
	u_int icol;

	icol = palette_getTrueColor32(color);
#if __BYTE_ORDER == __BIG_ENDIAN
	b = 0xff & (icol >> 16); m = 0xff & (icol >> 8); t = 0xff & icol;
#elif __BYTE_ORDER == __LITTLE_ENDIAN
	t = 0xff & (icol >> 16); m = 0xff & (icol >> 8); b = 0xff & icol;
#else
	#error FIXME : No endianness ?
#endif

	for (y = sy; y < sy + ly; y++) {
		for (x = 0; x < (lx * 3); x += 3) {
			p->smem[y * p->bytePerLine + sx * 3 + x] ^= t;
			p->smem[y * p->bytePerLine + sx * 3 + x + 1] ^= m;
			p->smem[y * p->bytePerLine + sx * 3 + x + 2] ^= b;
		}
	}
}
#endif /* ENABLE_24BPP */

#ifdef ENABLE_32BPP
/* 32 bpp */
void fbaccess_fillRect_32bpp(TFrameBuffer *p, u_int sx, u_int sy,
			     u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_int icol;
	u_int *d;

	icol = palette_getTrueColor32(color);
	for (y = sy; y < sy + ly; y++) {
		d = (u_int *)(p->smem + y * p->bytePerLine + sx * 4);
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
#ifdef ENABLE_SPLASH_SCREEN
			if (gSplash.enable && icol == 0 &&
			    SPLASH_INRANGE(gSplash.width, gSplash.height,
					   sx + x, y))
				*d++ = SPLASH_GET_PIXEL(gSplash, sx + x, y);
			else
				*d++ = icol;
#else
			*d++ = icol;
#endif
		}
	}
}

void fbaccess_clearAll_32bpp(TFrameBuffer *p, u_char color)
{
	u_int lp;
	u_int *d;
	u_int icol;

	d = (u_int *)(p->smem);
	icol = palette_getTrueColor32(color);
	for (lp = 0; lp < ((p->slen) / 2); lp++)
		d[lp] = icol;
}

void fbaccess_overlay_32bpp(TFrameBuffer *p, u_int xd, u_int yd,
			    const u_char *ps, u_int lx, u_int ly,
			    u_int gap, u_char color)
{
	u_int y;
	u_int *wp;
	const u_char *tps;
	u_int i;
	u_char sb;
	u_int icol;

	icol = palette_getTrueColor32(color);
	for (y = yd; y < yd + ly; y++) {
		tps = ps;
		wp = (u_int *)(p->smem + y * p->bytePerLine + xd * 4);
		for (i = lx; i >= 8; i -= 8) {
			sb = *tps++;
			if (sb & 0x80) wp[0] = icol;
			if (sb & 0x40) wp[1] = icol;
			if (sb & 0x20) wp[2] = icol;
			if (sb & 0x10) wp[3] = icol;
			if (sb & 0x08) wp[4] = icol;
			if (sb & 0x04) wp[5] = icol;
			if (sb & 0x02) wp[6] = icol;
			if (sb & 0x01) wp[7] = icol;
			wp += 8;
		}
		if (i != 0) {
			sb = *tps++;
			switch (i) {
			case 7: if (sb & 0x02) wp[6] = icol;
				/* FALLTHROUGH */
			case 6: if (sb & 0x04) wp[5] = icol;
				/* FALLTHROUGH */
			case 5: if (sb & 0x08) wp[4] = icol;
				/* FALLTHROUGH */
			case 4: if (sb & 0x10) wp[3] = icol;
				/* FALLTHROUGH */
			case 3: if (sb & 0x20) wp[2] = icol;
				/* FALLTHROUGH */
			case 2: if (sb & 0x40) wp[1] = icol;
				/* FALLTHROUGH */
			case 1: if (sb & 0x80) wp[0] = icol;
				/* FALLTHROUGH */
			default: break;
			}
		}
		ps += gap;
	}
}

void fbaccess_reverse_32bpp(TFrameBuffer *p, u_int sx, u_int sy,
			    u_int lx, u_int ly, u_char color)
{
	u_int x, y;
	u_int *d;
	u_int icol;

	icol = palette_getTrueColor32(color);
	for (y = sy; y < sy + ly; y++) {
		d = (u_int *)(p->smem + y * p->bytePerLine + sx * 4);
		for (x = 0; x < lx; x++) {
			if (sx + x >= p->width) break;
			d[x] ^= icol;
		}
	}
}
#endif /* ENABLE_32BPP */

