/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM/ME - J framebuffer terminal/Multilingual Enhancement -
 * Copyright (C) 2003 Fumitoshi UKAI <ukai@debian.or.jp>
 * Copyright (C) 1999 Noritoshi MASUICHI <nmasu@ma3.justnet.ne.jp>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

#include "csv.h"
#include "font.h"
#include "framebuffer.h"
#include "pcf.h"
#include "palette.h"
#include "picofont.h"
#include "privilege.h"
#include "utilities.h"

u_int gFontsWidth = 0;
u_int gFontsHeight = 0;

static u_char fontsDefaultGlyph[] = { 0x80 };
static u_char fontsDefaultDGlyph[] = { 0xc0 };
static pid_t child;

static void finalize(void);
static void finalizer(void);
static void configFontset(TCapValue *capValue);
static void configEffect(const char *config);
static void configShadowColor(const char *config);
static TFont *getFontByName(const char *name);
static const u_char *defaultGlyph(TFont *p, u_short chlw, u_int *width);
static const u_char *standardGlyph(TFont *p, u_short chlw, u_int *width);
static void childHandler(int signum);
static FILE *openStream(const char *path);
static void setupPCFFont(TFont *p, const char *path, FONTSET_HALF fhalf);
static void setupFontAlias(TFont *dst, TFont *src, FONTSET_HALF fhalf);
static void setupGlyphSize(void);

#define FMACRO_94__FONT(final, aline, fontname) { \
		.conv = defaultGlyph, \
		.name = fontname, \
		.width = 1, \
		.height = 1, \
		.fsignature = TFONT_FT_SINGLE | TFONT_FT_94CHAR | final, \
		.fhalf = aline,	\
		.alias = false, \
		.glyph = NULL, \
		.glyph_width = NULL, \
		.dglyph = fontsDefaultGlyph, \
		.bitmap = NULL,	\
		.colf = 0xff, \
		.coll = 0x00, \
		.rowf = 0xff, \
		.rowl = 0x00, \
		.colspan = 0x00, \
		.bytew = 2, \
		.bytec = 1 \
			 }

#define FMACRO_96__FONT(final, aline, fontname) { \
		.conv = defaultGlyph, \
		.name = fontname, \
		.width = 1, \
		.height = 1, \
		.fsignature = TFONT_FT_SINGLE | TFONT_FT_96CHAR | final, \
		.fhalf = aline,	\
		.alias = false, \
		.glyph = NULL, \
		.glyph_width = NULL, \
		.dglyph = fontsDefaultGlyph, \
		.bitmap = NULL,	\
		.colf = 0xff, \
		.coll = 0x00, \
		.rowf = 0xff, \
		.rowl = 0x00, \
		.colspan = 0x00, \
		.bytew = 2, \
		.bytec = 1 \
			 }

#define FMACRO_94N_FONT(final, aline, fontname) { \
		.conv = defaultGlyph, \
		.name = fontname, \
		.width = 2, \
		.height = 1, \
		.fsignature = TFONT_FT_DOUBLE | TFONT_FT_94CHAR | final, \
		.fhalf = aline,	\
		.alias = false, \
		.glyph = NULL, \
		.glyph_width = NULL, \
		.dglyph = fontsDefaultDGlyph,	\
		.bitmap = NULL,	\
		.colf = 0xff, \
		.coll = 0x00, \
		.rowf = 0xff, \
		.rowl = 0x00, \
		.colspan = 0x00, \
		.bytew = 2, \
		.bytec = 2 \
			 }

TFont gFont[] = {
	FMACRO_94__FONT(0x40, FH_LEFT,	"iso646-1973irv"),
	/* CJK Character sets */
	FMACRO_94N_FONT(0x40, FH_LEFT,	"jisc6226-1978"),
	FMACRO_94N_FONT(0x41, FH_LEFT,	"gb2312-80"),
	FMACRO_94N_FONT(0x42, FH_LEFT,	"jisx0208-1983"),
#if 0
	FMACRO_94N_FONT(0x42, FH_LEFT,	"jisx0208-1990"),
#endif
	FMACRO_94N_FONT(0x43, FH_LEFT,	"ksc5601-1987"),
	FMACRO_94N_FONT(0x44, FH_LEFT,	"jisx0212-1990"),
	FMACRO_94N_FONT(0x45, FH_LEFT,	"ccitt-extended-gb" /*aka iso-ir-165*/),
	FMACRO_94N_FONT(0x47, FH_LEFT,	"cns11643-1992.1"),
	FMACRO_94N_FONT(0x48, FH_LEFT,	"cns11643-1992.2"),
	FMACRO_94N_FONT(0x49, FH_LEFT,	"cns11643-1992.3"),
	FMACRO_94N_FONT(0x4a, FH_LEFT,	"cns11643-1992.4"),
	FMACRO_94N_FONT(0x4b, FH_LEFT,	"cns11643-1992.5"),
	FMACRO_94N_FONT(0x4c, FH_LEFT,	"cns11643-1992.6"),
	FMACRO_94N_FONT(0x4d, FH_LEFT,	"cns11643-1992.7"),
	/* Domestic ISO 646 Character sets */
	FMACRO_94__FONT(0x41, FH_LEFT,	"bs4730"),
	FMACRO_94__FONT(0x42, FH_LEFT,	"ansix3.4-1968" /* aka ASCII */),
	FMACRO_94__FONT(0x43, FH_LEFT,	"nats1-finland-sweden"),
	FMACRO_94__FONT(0x47, FH_LEFT,	"sen850200b"),
	FMACRO_94__FONT(0x48, FH_LEFT,	"sen850200c"),
	FMACRO_94__FONT(0x4a, FH_LEFT,	"jisc6220-1969roman"),
	FMACRO_94__FONT(0x59, FH_LEFT,	"uni0204-70"),
	FMACRO_94__FONT(0x4c, FH_LEFT,	"olivetti-portuguese"),
	FMACRO_94__FONT(0x5a, FH_LEFT,	"olivetti-spanish"),
	FMACRO_94__FONT(0x4b, FH_LEFT,	"din66003"),
	FMACRO_94__FONT(0x52, FH_LEFT,	"nfz62-010-1973"),
	FMACRO_94__FONT(0x54, FH_LEFT,	"gb1988-80"),
	FMACRO_94__FONT(0x60, FH_LEFT,	"ns4551-1"),
	FMACRO_94__FONT(0x61, FH_LEFT,	"ns4551-2"),
	FMACRO_94__FONT(0x66, FH_LEFT,	"nfz62-010-1982"),
	FMACRO_94__FONT(0x67, FH_LEFT,	"ibm-portuguese"),
	FMACRO_94__FONT(0x68, FH_LEFT,	"ibm-spanish"),
	FMACRO_94__FONT(0x69, FH_LEFT,	"msz7795.3"),
	FMACRO_94__FONT(0x6e, FH_LEFT,	"jisc6229-1984ocr-b"),
	FMACRO_94__FONT(0x75, FH_LEFT,	"ccitt-t.61-1"),
	FMACRO_94__FONT(0x77, FH_LEFT,	"csaz243.4-1985alt1-1"),
	FMACRO_94__FONT(0x78, FH_LEFT,	"csaz243.4-1985alt1-2"),
	FMACRO_94__FONT(0x7a, FH_LEFT,	"jusi.b1.002"),
	/* Domestic ISO 646 Character sets with I-oct */
#if 0
	FMACRO_94__FONT(0x41, FH_LEFT,	"nc99-10:81"),
	FMACRO_94__FONT(0x42, FH_LEFT,	"iso646-1992invariants"),
#endif
	/* Quasi ISO 646 Character sets */
	FMACRO_94__FONT(0x45, FH_LEFT,	"nats1-denmark-norway"),
	FMACRO_94__FONT(0x55, FH_LEFT,	"honeywell-bull-latin-greek"),
	FMACRO_94__FONT(0x56, FH_LEFT,	"british-post-office-teletext"),
	FMACRO_94__FONT(0x57, FH_LEFT,	"inis-irv-subset"),
	/* Greek Character sets */
	FMACRO_94__FONT(0x5b, FH_LEFT,	"olivetti-greek"),
	FMACRO_94__FONT(0x5c, FH_LEFT,	"olivetti-latin-greek"),
	FMACRO_94__FONT(0x58, FH_LEFT,	"iso5428-1974"),
	FMACRO_94__FONT(0x53, FH_LEFT,	"iso5428-1980"),
	/* Greek Character sets  with I-oct */
#if 0
	FMACRO_94__FONT(0x40, FH_LEFT,	"ccitt-greek1"),
#endif
	/* Cyrillic Character sets */
	FMACRO_94__FONT(0x4e, FH_LEFT,	"iso5427-1981"),
	FMACRO_94__FONT(0x5e, FH_LEFT,	"inis-cyrillic-extension"),
	FMACRO_94__FONT(0x51, FH_LEFT,	"iso5427-1981extension"),
	FMACRO_96__FONT(0x40, FH_LEFT,	"ecma-94cyrillic"),
	FMACRO_94__FONT(0x7b, FH_LEFT,	"jusi.b1.003"),
	FMACRO_94__FONT(0x7d, FH_LEFT,	"jusi.b1.004"),
	FMACRO_96__FONT(0x4f, FH_LEFT,	"stsev358-88"),
	/* ISO 8859 variants */
	FMACRO_96__FONT(0x41, FH_RIGHT, "iso8859.1-1987"),
	FMACRO_96__FONT(0x42, FH_RIGHT, "iso8859.2-1987"),
	FMACRO_96__FONT(0x43, FH_RIGHT, "iso8859.3-1988"),
	FMACRO_96__FONT(0x44, FH_RIGHT, "iso8859.4-1988"),
	FMACRO_96__FONT(0x4c, FH_RIGHT, "iso8859.5-1988"),
	FMACRO_96__FONT(0x47, FH_RIGHT, "iso8859.6-1987"),
	FMACRO_96__FONT(0x46, FH_RIGHT, "iso8859.7-1987"),
	FMACRO_96__FONT(0x48, FH_RIGHT, "iso8859.8-1988"),
	FMACRO_96__FONT(0x4d, FH_RIGHT, "iso8859.9-1989"),
	FMACRO_96__FONT(0x56, FH_RIGHT, "iso8859.10-1992"),
	FMACRO_96__FONT(0x54, FH_RIGHT, "iso8859.11-2001"),
	FMACRO_96__FONT(0x59, FH_RIGHT, "iso8859.13-1998"),
	FMACRO_96__FONT(0x5f, FH_RIGHT, "iso8859.14-1998"),
	FMACRO_96__FONT(0x62, FH_RIGHT, "iso8859.15-1999"),
	FMACRO_96__FONT(0x66, FH_RIGHT, "iso8859.16-2001"),
	/* CCITT T.101 Mosaic Charactor Sets */
	FMACRO_94__FONT(0x79, FH_LEFT,	"ccitt-mosaic-1"),
	FMACRO_96__FONT(0x7d, FH_LEFT,	"ccitt-mosaic-sup1"),
	FMACRO_94__FONT(0x63, FH_LEFT,	"ccitt-mosaic-sup2"),
	FMACRO_94__FONT(0x64, FH_LEFT,	"ccitt-mosaic-sup3"),
	/* Japano */
	FMACRO_94__FONT(0x49, FH_LEFT,	"jisc6220-1969kana"),
	/* JISX0213 */
	FMACRO_94N_FONT(0x4f, FH_LEFT,	"jisx0213-2000-1"),
	FMACRO_94N_FONT(0x50, FH_LEFT,	"jisx0213-2000-2"),
	/* Linux Private */
	FMACRO_94__FONT(0x30, FH_LEFT,	"vt100-graphics"),
#ifdef ENABLE_UTF8
	/* Unicode 2.0 */
	{
		.conv = defaultGlyph,
		.name = "iso10646.1",
		.width = 1,
		.height = 1,
		.fsignature = TFONT_FT_OTHER,
		.fhalf = FH_UNI,
		.alias = false,
		.glyph = NULL,
		.glyph_width = NULL,
		.dglyph = fontsDefaultDGlyph,
		.bitmap = NULL,
		.colf = 0xff,
		.coll = 0x00,
		.rowf = 0xff,
		.rowl = 0x00,
		.colspan = 0x00,
		.bytew = 1,
		.bytec = 1
	},
#endif
	{
		.conv = defaultGlyph,
		.name = NULL,
		.width = 1,
		.height = 1,
		.fsignature = 0x00000000,
		.fhalf = FH_LEFT,
		.alias = false,
		.glyph = NULL,
		.glyph_width = NULL,
		.dglyph = fontsDefaultDGlyph,
		.bitmap = NULL,
		.colf = 0xff,
		.coll = 0x00,
		.rowf = 0xff,
		.rowl = 0x00,
		.colspan = 0x00,
		.bytew = 1,
		.bytec = 1
	}
};

typedef enum {
	FONT_EFFECT_OFF,
	FONT_EFFECT_SHADOW
} FONT_EFFECT;

#define DEFAULT_FONT_SHADOW_COLOR (0) /* Black */

static struct {
	FONT_EFFECT effect;
	u_char shadowColor;
} font = {
	FONT_EFFECT_OFF,
	DEFAULT_FONT_SHADOW_COLOR
};

static bool initialized;

void font_initialize(void)
{
	assert(!initialized);

	atexit(finalizer);
	initialized = true;
}

static void finalize(void)
{
	TFont *p;

	assert(initialized);

	for (p = gFont; p->name != NULL; p++) {
		if (!p->alias && font_isLoaded(p)) {
			p->conv = defaultGlyph;
			if (p->glyph != NULL) {
				free(p->glyph);
				p->glyph = NULL;
			}
			if (p->glyph_width != NULL) {
				free(p->glyph_width);
				p->glyph_width = NULL;
			}
			if (p->dglyph != NULL) {
				free(p->dglyph);
				p->dglyph = NULL;
			}
			if (p->bitmap != NULL) {
				free(p->bitmap);
				p->bitmap = NULL;
			}
		}
	}
	initialized = false;
}

static void finalizer(void)
{
	finalize();
}

static void configFontset(TCapValue *capValue)
{
	struct sigaction act, oldact;
	const char *name, *type, *side, *path;
	TFont *dst, *src;
	TCapValue *p;
	TCsv csv;
	FONTSET_HALF fhalf;
	int i;

	assert(initialized);
	assert(capValue != NULL);

	bzero(&act, sizeof(act));
	act.sa_handler = childHandler;
	act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
	sigaction(SIGCHLD, &act, &oldact);

	for (p = capValue; p != NULL; p = p->next) {
		csv_initialize(&csv, p->value);
		if (csv.tokens != 4) {
			warnx("FONT : Skipped (BAD FORMAT)");
			csv_finalize(&csv);
			continue;
		}
		name = csv_getToken(&csv);
		assert(name != NULL);
		type = csv_getToken(&csv);
		assert(type != NULL);
		side = csv_getToken(&csv);
		assert(side != NULL);
		path = csv_getToken(&csv);
		assert(path != NULL);
		fprintf(stderr, "FONT : [%s]/%s/%s://%s/\n",
			name, type, side, path);
		if ((dst = getFontByName(name)) == NULL) {
			warnx("FONT : Skipped (Unknown FONTSET=`%s')", name);
			csv_finalize(&csv);
			continue;
		}
		i = lookup(side, "L", "R", "U", (char *)NULL);
		switch (i) {
		case 0:
			fhalf = FH_LEFT;
			break;
		case 1:
			fhalf = FH_RIGHT;
			break;
		case 2:
			fhalf = FH_UNI;
			break;
		default:
			warnx("FONT : Skipped (BAD FORMAT)");
			csv_finalize(&csv);
			continue;
		}
		i = lookup(type, "pcf", "alias", (char *)NULL);
		switch (i) {
		case 0:
			setupPCFFont(dst, path, fhalf);
			break;
		case 1:
			if ((src = getFontByName(path)) == NULL) {
				warnx("FONT : Skipped (Unknown ALIAS FONTSET=`%s')", path);
				csv_finalize(&csv);
				continue;
			}
			setupFontAlias(dst, src, fhalf);
			break;
		default:
			warnx("FONT : Skipped (BAD FORMAT)");
			csv_finalize(&csv);
			continue;
		}
		csv_finalize(&csv);
	}

	sigaction(SIGCHLD, &oldact, NULL);

	setupGlyphSize();

	if (!font_isLoaded(&(gFont[0]))) {
#ifdef ENABLE_UTF8
		bool loaded;
		loaded = false;
		for (i = 1; gFont[i].name != NULL; i++) {
			if (gFont[i].fhalf == FH_UNI) {
				loaded = font_isLoaded(&(gFont[i]));
				break;
			}
		}
		if (!loaded)
			errx(1, "FONT : ISO8859/ISO10646 not loaded.");
#else
		errx(1, "FONT : ISO8859 not loaded.");
#endif
	}
}

static void configEffect(const char *config)
{
	bool found;
	int i;

	static const struct {
		const char *key;
		const FONT_EFFECT effect;
	} list[] = {
		{ "Shadow", FONT_EFFECT_SHADOW },
		{ "Off",    FONT_EFFECT_OFF    },
		{ NULL,     FONT_EFFECT_OFF    }
	};

	assert(initialized);

	font.effect = FONT_EFFECT_OFF;
	if (config != NULL) {
		found = false;
		for (i = 0; list[i].key != NULL; i++) {
			if (strcasecmp(list[i].key, config) == 0) {
				font.effect = list[i].effect;
				found = true;
				break;
			}
		}
		if (!found)
			warnx("Invalid font effect: %s", config);
	}
}

static void configShadowColor(const char *config)
{
	int color;

	assert(initialized);

	if (config != NULL) {
		color = atoi(config);
		if (color >= 0 && color <= 15)
			font.shadowColor = palette_ansiToVGA(color);
		else
			warnx("Invalid font shadow color: %s", config);
	}
}

void font_configure(TCaps *caps)
{
	TCapability *capability;
	const char *config;

	assert(initialized);
	assert(caps != NULL);

	capability = caps_find(caps, "fontset");
	if (capability == NULL || capability->values == NULL)
		errx(1, "No font specified.");
	configFontset(capability->values);
	config = caps_findFirst(caps, "font.effect");
	configEffect(config);
	config = caps_findFirst(caps, "font.shadow.color");
	configShadowColor(config);
}

void font_showList(FILE *stream)
{
	TFont *p;
	char c;
	u_int n, u, d, f1, f2;

	/* not require any initialize */
	assert(stream != NULL);

	for (p = gFont; p->name != NULL; p++) {
		n = p->fsignature;
		c = n & 255;
		u = (n >> 4) & 15;
		d = (n >> 0) & 15;
		f1 = n & TFONT_FT_DOUBLE;
		f2 = n & TFONT_FT_96CHAR;
		if (p->fhalf == FH_UNI)
			fprintf(stream, " [--/--(-) uni ] %s\n", p->name);
		else {
			fprintf(stream, " [%02u/%02u(%c) %s%s] %s\n",
				u, d, c,
				f2 ? "96" : "94",
				f1 ? "^N" : "  ",
				p->name);
		}
	}
}

int font_getIndexBySignature(const u_int fsignature)
{
	TFont *p;
	int i;

	/* not require any initialize */

	for (p = gFont, i = 0; p->name != NULL; p++, i++)
		if (p->fsignature == fsignature)
			return i;
	return -1;
}

int font_getIndexByName(const char *name)
{
	TFont *p;
	int i;

	/* not require any initialize */
	assert(name != NULL);

	for (p = gFont, i = 0; p->name != NULL; p++, i++)
		if (strncasecmp(p->name, name, strlen(p->name)) == 0)
			return i;
	return -1;
}

static TFont *getFontByName(const char *name)
{
	TFont *p;

	assert(initialized);
	assert(name != NULL);

	for (p = gFont; p->name != NULL; p++)
		if (strncasecmp(p->name, name, strlen(p->name)) == 0)
			return p;
	return NULL;
}

static const u_char *defaultGlyph(TFont *p, u_short chlw, u_int *width)
{
	static u_char glyph[PICOFONT_HEIGHT * 2];

	u_char *cp;
	u_int i;
	u_int a;
	u_int b;

	assert(initialized);
	assert(p != NULL);
	assert(width != NULL);

	cp = glyph;
	*width = p->width;
	*cp++ = 0x80;
	*cp++ = 0x00;
	for (i = 0; i < 16; i += 4) {
		b = (chlw >> (12 - i)) & 0xf;
		b <<= 1;
		a = (chlw >> (28 - i)) & 0xf;
		a <<= 1;
		*cp++ = 0x80 | gPicoFontLeft[b++];
		*cp++ = gPicoFontLeft[a++];
		*cp++ = 0x80 | gPicoFontLeft[b];
		*cp++ = gPicoFontLeft[a];
		*cp++ = 0x80;
		*cp++ = 0x00;
	}
	*cp++ = 0x80;
	*cp++ = 0x00;
	if (p->bytec == 2) {
		*cp++ = 0xff;
		*cp = 0xfe;
	} else {
		*cp++ = 0xfe;
		*cp = 0x00;
	}
	return glyph;
}

static const u_char *standardGlyph(TFont *p, u_short chlw, u_int *width)
{
	u_char b1, b2;
	u_int i;

	assert(initialized);
	assert(p != NULL);
	assert(width != NULL);

	b1 = chlw & 0xff;
	b2 = (chlw >> 8) & 0xff;
	*width = p->width;
	if (b1 < p->colf || p->coll < b1 || b2 < p->rowf || p->rowl < b2)
		return p->dglyph;
	else {
		i = (b1 - p->colf) + (b2 - p->rowf) * p->colspan;
		if (p->glyph_width != NULL)
			*width = p->glyph_width[i];
		return p->glyph[i];
	}
}

static void childHandler(int signum)
{
	int status, errsv;
	pid_t pid;

	errsv = errno;
	pid = waitpid(child, &status, WNOHANG);
	if (pid == (pid_t)-1 && errno != ECHILD)
		err(1, "waitpid()");
	if (pid == child || (pid == (pid_t)-1 && errno == ECHILD)) {
		if (WIFSIGNALED(status))
			errx(1, "Child process died with signal %s",
			     sys_siglist[WTERMSIG(status)]);
		if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
			errx(1, "Child process exited with status %d",
			     WEXITSTATUS(status));
	}
	errno = errsv;
}

static FILE *openStream(const char *path)
{
	int i, fd, filedes[2];

	assert(initialized);
	assert(path != NULL);

	if ((strlen(path) > 3 && strrcmp(path, ".gz") == 0) ||
	    (strlen(path) > 2 && strrcmp(path, ".Z") == 0)) {
		if (pipe(filedes) == -1)
			return NULL;
		fflush(stdout);
		fflush(stderr);
		child = fork();
		if (child == (pid_t)-1)
			err(1, "fork()");
		else if (child == (pid_t)0) {
			/* Child process */
			privilege_drop();
			if ((fd = open(path, O_RDONLY)) == -1) {
				close(filedes[0]);
				close(filedes[1]);
				exit(EXIT_FAILURE);
				/* NOTREACHED */
			}
			close(filedes[0]);
			close(STDIN_FILENO);
			close(STDOUT_FILENO);
			dup2(fd, STDIN_FILENO);
			dup2(filedes[1], STDOUT_FILENO);
			for (i = getdtablesize(); i > STDERR_FILENO; i--)
				close(i);
#ifndef GUNZIP_PATH
	#error GUNZIP_PATH is not defined.
#else
			execl(GUNZIP_PATH, GUNZIP_PATH, "-c", (char *)NULL);
#endif
			exit(EXIT_FAILURE);
			/* NOTREACHED */
		} else {
			/* Parent process */
			close(filedes[1]);
			return fdopen(filedes[0], "r");
		}
	} else
		return fopen(path, "r");
	return NULL;
}

static void setupPCFFont(TFont *p, const char *path, FONTSET_HALF fhalf)
{
	FILE *stream;
	TPcf pcf;

	assert(initialized);
	assert(p != NULL);
	assert(path != NULL);

	pcf_initialize(&pcf);
	if (access(path, R_OK) != 0) {
		warnx("PCF : READ ACCESS ERROR : %s", path);
		return;
	}
	stream = openStream(path);
	if (stream == NULL) {
		warnx("PCF : CANNOTOPEN : %s", path);
		return;
	}
	tpcf_load(&pcf, stream);
	fclose(stream);
	tpcf_as_tfont(&pcf, p);
	p->fhalf = fhalf;
	p->conv = standardGlyph;
	pcf_finalize(&pcf);
}

static void setupFontAlias(TFont *dst, TFont *src, FONTSET_HALF fhalf)
{
	assert(initialized);
	assert(dst != NULL);
	assert(src != NULL);

	dst->conv = src->conv;
	dst->width = src->width;
	dst->height = src->height;
	dst->fhalf = fhalf;
	dst->alias = true;
	dst->glyph = src->glyph;
	dst->glyph_width = src->glyph_width;
	dst->dglyph = src->dglyph;
	dst->bitmap = src->bitmap;
	dst->colf = src->colf;
	dst->coll = src->coll;
	dst->rowf = src->rowf;
	dst->rowl = src->rowl;
	dst->colspan = src->colspan;
	dst->bytew = src->bytew;
	dst->bytec = src->bytec;
}

static void setupGlyphSize(void)
{
	TFont *p;
	u_int w;

	assert(initialized);

	for (p = gFont; p->name != NULL; p++) {
		if (gFontsHeight < p->height)
			gFontsHeight = p->height;
		w = p->width;
		w = (p->fsignature & TFONT_FT_DOUBLE) ? (w + 1) / 2 : w;
		if (gFontsWidth < w)
			gFontsWidth = w;
	}
	if (PICOFONT_WIDTH <= gFontsWidth && PICOFONT_HEIGHT <= gFontsHeight) {
		for (p = gFont; p->name != NULL; p++) {
			if (p->conv == defaultGlyph) {
				p->height = PICOFONT_HEIGHT;
				p->width = PICOFONT_WIDTH * p->width;
			}
		}
	}
}

bool font_isLoaded(TFont *p)
{
	assert(initialized);
	assert(p != NULL);

	return p->conv != defaultGlyph;
}

void font_draw(TFont *p, u_short chlw, u_char fcol, u_char bcol, u_int x, u_int y, u_int w)
{
	const u_char *glyph;
	u_int fx, fy, fw, fh, gw, sw, sh;

	assert(initialized);
	assert(p != NULL);

	fx = gFontsWidth * x;
	fy = gFontsHeight * y;
	fw = gFontsWidth * w;
	fh = gFontsHeight;
	gFramebuffer.cap.fill(&gFramebuffer, fx, fy, fw, fh, bcol);
	if (chlw == 0)
		return;
	glyph = p->conv(p, chlw, &gw);
	if (font.effect == FONT_EFFECT_SHADOW &&
	    gw > 0 && p->height > 0 && bcol == 0) {
		sw = (fw > gw) ? gw : gw - 1;
		sh = (fh > p->height) ? p->height : p->height - 1;
		gFramebuffer.cap.overlay(&gFramebuffer,
					 fx + 1,
					 fy + 1,
					 glyph,
					 sw,
					 sh,
					 p->bytew,
					 font.shadowColor);
	}
	gFramebuffer.cap.overlay(&gFramebuffer,
				 fx,
				 fy,
				 glyph,
				 gw,
				 p->height,
				 p->bytew,
				 fcol);
}

