/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * KON2 - Kanji ON Console -
 * Copyright (C) 1992,1993 Atusi MAEDA <mad@math.keio.ac.jp>
 * Copyright (C) 1992-1996 Takashi MANABE <manabe@papilio.tutics.tut.ac.jp>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#if defined (__NetBSD__) || defined (__OpenBSD__)
#include <time.h>
#include <dev/wscons/wsconsio.h>
#endif

#include "clipboard.h"
#include "font.h"
#include "keyboard.h"
#include "mouse.h"
#include "privilege.h"
#ifdef ENABLE_SCROLLBACK
#include "scrollback.h"
#endif
#include "utilities.h"
#include "vtermlow.h"

#define MAX_PACKET_SIZE     (5)
#define BUF_SIZE	    (MAX_PACKET_SIZE * 100)

#define MOUSE_LIFETIME      (55) /* 5.5 sec. */

#define MOUSE_BUTTON_LEFT   (4)
#define MOUSE_BUTTON_MIDDLE (2)
#define MOUSE_BUTTON_RIGHT  (1)

static struct {
	int x, y;
	int sx, sy;
	int dx, dy;
	char status;
} mouseInfo;

typedef enum {
	MOUSE_TYPE_MICROSOFT,
	MOUSE_TYPE_MOUSESYSTEMS,
	MOUSE_TYPE_BUSMOUSE,
	MOUSE_TYPE_MMSERIES,
	MOUSE_TYPE_LOGITECH,
	MOUSE_TYPE_PS2,
#if 0 /* TODO: maybe does not work */
	MOUSE_TYPE_IMPS2,
#endif
#if defined (__NetBSD__) || defined (__OpenBSD__)
	MOUSE_TYPE_WSMOUSE,
#endif
	MOUSE_TYPE_NONE
} MOUSE_TYPE;

static struct {
	int fd;
	MOUSE_TYPE type;
	int cFlag;
	int headMask;
	int headId;
	int dataMask;
	int pkMax;
	int baudRate;
	char *device;
	int pasteButton;
	struct timeval timer;
	bool enable;
} mouse = {
	-1,
	MOUSE_TYPE_NONE,
	0,
	0,
	0,
	0,
	0,
	B1200,
	NULL,
	MOUSE_BUTTON_MIDDLE,
	{ 0, 0 },
	false
};

static bool initialized;

static void finalize(void);
static void finalizer(void);
static void configBaudRate(const char *config);
static void configDevice(const char *config);
static void configPasteButton(const char *config);
static void configType(const char *config);
static void setBaudRate(const int baudRate, const int cFlag);
static void resetTimer(void);
static void analyzePacket(TVterm *p, const u_char *packet);
#if defined (__NetBSD__) || defined (__OpenBSD__)
static void analyzeWSConsEvents(TVterm *p);
#endif
static void selection(TVterm *p, const char oldstatus);

void mouse_initialize(void)
{
	assert(!initialized);

	atexit(finalizer);
	initialized = true;
	clipboard_initialize();
}

static void finalize(void)
{
	assert(initialized);

	mouse_stop();
	if (mouse.device != NULL) {
		free(mouse.device);
		mouse.device = NULL;
	}
	mouse.enable = false;
	initialized = false;
}

static void finalizer(void)
{
	finalize();
}

static void configBaudRate(const char *config)
{
	assert(initialized);

	if (config != NULL)
		mouse.baudRate = atoi(config);
	switch (mouse.baudRate) {
	case 9600:
		mouse.baudRate = B9600;
		break;
	case 4800:
		mouse.baudRate = B4800;
		break;
	case 2400:
		mouse.baudRate = B2400;
		break;
	default:
		warnx("Invalid mouse baud rate %s", config);
		/* FALLTHROUGH */
	case 1200:
		mouse.baudRate = B1200;
		break;
	}
}

static void configDevice(const char *config)
{
	assert(initialized);

	if (config != NULL) {
		if (mouse.device)
			free(mouse.device);
		mouse.device = strdup(config);
		if (mouse.device == NULL)
			err(1, "strdup()");
	}
}

static void configPasteButton(const char *config)
{
	bool found;
	int i;

	static const struct {
		const char *key;
		const int pasteButton;
	} list[] = {
		{ "Middle", MOUSE_BUTTON_MIDDLE },
		{ "Right",  MOUSE_BUTTON_RIGHT  },
		{ NULL,     0                   }
	};

	assert(initialized);

	mouse.pasteButton = MOUSE_BUTTON_MIDDLE;
	if (config != NULL) {
		found = false;
		for (i = 0; list[i].key != NULL; i++) {
			if (strcasecmp(list[i].key, config) == 0) {
				mouse.pasteButton = list[i].pasteButton;
				found = true;
				break;
			}
		}
		if (!found)
			warnx("Invalid mouse paste button: %s", config);
	}
}

static void configType(const char *config)
{
	bool found;
	int i;

	static const struct {
		const char *name;
		const MOUSE_TYPE type;
		const int cFlag;
		const int headMask;
		const int headId;
		const int dataMask;
		const int pkMax;
		const bool enable;
	} list[] = {
		{
			"Microsoft", MOUSE_TYPE_MICROSOFT,
			(CS7 | CREAD | CLOCAL | HUPCL),
			0x40, 0x40, 0x40, 3, true
		},
		{
			"MouseSystems", MOUSE_TYPE_MOUSESYSTEMS,
			(CS8 | CSTOPB | CREAD | CLOCAL | HUPCL),
			0xf8, 0x80, 0x00, 5, true
		},
		{
			"BusMouse", MOUSE_TYPE_BUSMOUSE,
			0,
			0xf8, 0x80, 0x00, 5, true
		},
		{
			"MmSeries", MOUSE_TYPE_MMSERIES,
			(CS8 | PARENB | PARODD | CREAD | CLOCAL | HUPCL),
			0xe0, 0x80, 0x80, 3, true
		},
		{
			"Logitech", MOUSE_TYPE_LOGITECH,
			(CS8 | CSTOPB | CREAD | CLOCAL | HUPCL),
			0xe0, 0x80, 0x80, 3, true
		},
		{
			"PS2", MOUSE_TYPE_PS2,
			(CS8 | CREAD | CLOCAL | HUPCL),
			0xcc, 0x08, 0x00, 3, true
		},
#if 0 /* TODO: maybe does not work */
		{
			"IMPS2", MOUSE_TYPE_IMPS2,
			(CS8 | CREAD | CLOCAL | HUPCL),
			0xcc, 0x08, 0x00, 4, true
		},
#endif
#if defined (__NetBSD__) || defined (__OpenBSD__)
		{
			"WSMouse", MOUSE_TYPE_WSMOUSE,
			0,
			0, 0, 0, 0, true
		},
#endif
		{
			"None", MOUSE_TYPE_NONE,
			0,
			0, 0, 0, 0, false
		},
		{
			NULL, MOUSE_TYPE_NONE,
			0,
			0, 0, 0, 0, false
		}
	};

	assert(initialized);

	mouse.type = MOUSE_TYPE_NONE;
	mouse.enable = false;
	if (config != NULL) {
		found = false;
		for (i = 0; list[i].name != NULL; i++) {
			if (strcasecmp(list[i].name, config) == 0) {
				mouse.type     = list[i].type;
				mouse.cFlag    = list[i].cFlag;
				mouse.headMask = list[i].headMask;
				mouse.headId   = list[i].headId;
				mouse.dataMask = list[i].dataMask;
				mouse.pkMax    = list[i].pkMax;
				mouse.enable   = list[i].enable;
				found = true;
				break;
			}
		}
		if (!found)
			warnx("Invalid mouse type: %s", config);
	}
}

void mouse_configure(TCaps *caps)
{
	const char *config;

	assert(initialized);
	assert(caps != NULL);

	config = caps_findFirst(caps, "mouse.type");
	configType(config);
	if (mouse.type != MOUSE_TYPE_NONE) {
		if (mouse.type != MOUSE_TYPE_BUSMOUSE) {
			config = caps_findFirst(caps, "mouse.baudRate");
			configBaudRate(config);
		}
		config = caps_findFirst(caps, "mouse.device");
		configDevice(config);
		config = caps_findFirst(caps, "mouse.pasteButton");
		configPasteButton(config);
	}
}

static void setBaudRate(const int baudRate, const int cFlag)
{
	struct termios mio;
	const char *cf;

	assert(initialized);

	if (mouse.fd == -1)
		return;
	tcgetattr(mouse.fd, &mio);
	mio.c_iflag = IGNBRK | IGNPAR;
	mio.c_oflag = 0;
	mio.c_lflag = 0;
#if defined (__linux__)
	mio.c_line = 0;
#endif
	mio.c_cc[VTIME] = 0;
	mio.c_cc[VMIN] = 1;
	mio.c_cflag = cFlag;
	cfsetispeed(&mio, baudRate);
	cfsetospeed(&mio, baudRate);
	tcsetattr(mouse.fd, TCSAFLUSH, &mio);
	switch (mouse.baudRate) {
	case B9600: cf = "*q"; break;
	case B4800: cf = "*p"; break;
	case B2400: cf = "*o"; break;
	default:
		/* FALLTHROUGH */
	case B1200: cf = "*n"; break;
	}
	mio.c_cflag = cFlag;
	cfsetispeed(&mio, mouse.baudRate);
	cfsetospeed(&mio, mouse.baudRate);
	write_wrapper(mouse.fd, cf, 2);
	usleep(100000);
	tcsetattr(mouse.fd, TCSAFLUSH, &mio);
}

int mouse_getFd(void)
{
	assert(initialized);

	return mouse.fd;
}

bool mouse_isEnable(void)
{
	assert(initialized);

	return mouse.enable;
}

static void resetTimer(void)
{
	assert(initialized);

	gettimeofday(&mouse.timer, NULL);
}

bool mouse_isTimeout(void)
{
	struct timeval now, diff;
	u_int elapsedTime;

	assert(initialized);

	timerclear(&diff);
	gettimeofday(&now, NULL);
	timersub(&now, &mouse.timer, &diff);
	elapsedTime = diff.tv_sec * 10 + diff.tv_usec / 100000; /* 0.1 sec. */
	return elapsedTime >= MOUSE_LIFETIME;
}

int mouse_getX(void)
{
	assert(initialized);

	return mouseInfo.x;
}

int mouse_getY(void)
{
	assert(initialized);

	return mouseInfo.y;
}

#if 0 /* TODO: maybe does not work */
#define SAMPLERATE (0xf3)

static void setupIntelliMouse(void)
{
	int i;
	unsigned char sampleRates[3][2] = {
		{ SAMPLERATE, 200 },
		{ SAMPLERATE, 100 },
		{ SAMPLERATE,  80 }
	};

	assert(initialized);

	for (i = 0; i < 3; i++)
		write_wrapper(mouse.fd, sampleRates[i], 2);
}
#endif

void mouse_start(void)
{
	struct stat st;

	assert(initialized);

	mouse_stop();
	if (!mouse.enable)
		return;
	if (mouse.device == NULL) {
		mouse.enable = false;
		return;
	}
	privilege_on();
	mouse.fd = open(mouse.device, O_RDWR | O_NONBLOCK);
	privilege_off();
	if (mouse.fd == -1) {
		warnx("Could not open mouse device: %s", mouse.device);
		free(mouse.device);
		mouse.device = NULL;
		mouse.enable = false;
		return;
	}
	if (fstat(mouse.fd, &st) == -1)
		err(1, "fstat(%s)", mouse.device);
	if (!S_ISCHR(st.st_mode)) {
		warnx("%s is not a character device.", mouse.device);
		free(mouse.device);
		mouse.device = NULL;
		mouse.enable = false;
		return;
	}
	if (fcntl(mouse.fd, F_SETFD, FD_CLOEXEC) == -1)
		err(1, "fcntl(F_SETFD)");
#if 0 /* TODO: maybe does not work */
	if (mouse.type == MOUSE_TYPE_IMPS2)
		setupIntelliMouse();
#endif
	if (mouse.type != MOUSE_TYPE_BUSMOUSE) {
		setBaudRate(B9600, mouse.cFlag);
		setBaudRate(B4800, mouse.cFlag);
		setBaudRate(B2400, mouse.cFlag);
		setBaudRate(B1200, mouse.cFlag);
		if (mouse.type == MOUSE_TYPE_LOGITECH) {
			write_wrapper(mouse.fd, "S", 1);
			setBaudRate(mouse.baudRate,
				    CS8 | PARENB | PARODD | CREAD |
				    CLOCAL | HUPCL);
		}
		write_wrapper(mouse.fd, "Q", 1);
	}
}

void mouse_stop(void)
{
	assert(initialized);

	if (mouse.fd != -1) {
		close(mouse.fd);
		mouse.fd = -1;
	}
}

static void analyzePacket(TVterm *p, const u_char *packet)
{
	static char oldstatus;
	static int dx, dy;

	assert(initialized);
	assert(p != NULL);
	assert(packet != NULL);

	switch (mouse.type) {
	case MOUSE_TYPE_MICROSOFT:
		mouseInfo.status = ((packet[0] & 0x20) >> 3) |
				   ((packet[0] & 0x10) >> 4);
		dx += (char)(((packet[0] & 0x03) << 6) | (packet[1] & 0x3f));
		dy += (char)(((packet[0] & 0x0c) << 4) | (packet[2] & 0x3f));
		break;
	case MOUSE_TYPE_MOUSESYSTEMS:
		mouseInfo.status = (~packet[0]) & 0x07;
		dx += (char)(packet[1]) + (char)(packet[3]);
		dy += -((char)(packet[2]) + (char)(packet[4]));
		break;
	case MOUSE_TYPE_BUSMOUSE:
		mouseInfo.status = (~packet[0]) & 0x07;
		dx += (char)packet[1];
		dy += -(char)packet[2];
		break;
	case MOUSE_TYPE_MMSERIES:
		/* FALLTHROUGH */
	case MOUSE_TYPE_LOGITECH:
		mouseInfo.status = packet[0] & 0x07;
		dx += (packet[0] & 0x10) ? packet[1] : -packet[1];
		dy += (packet[0] & 0x08) ? -packet[2] : packet[2];
		break;
#if 0 /* TODO: maybe does not work */
	case MOUSE_TYPE_IMPS2:
		/* FALLTHROUGH */
#endif
	case MOUSE_TYPE_PS2:
		/* Middle button: IntelliMouse extension. */
		mouseInfo.status = ((packet[0] & 0x01) << 2) |
				   ((packet[0] & 0x02) >> 1) |
				   ((packet[0] & 0x04) >> 1);
		dx += (char)(packet[1]);
		dy -= (char)(packet[2]);
		break;
	case MOUSE_TYPE_NONE:
		/* FALLTHROUGH */
	default:
		return;
	}
	mouseInfo.dx = dx / (int)gFontsWidth;
	dx -= mouseInfo.dx * (int)gFontsWidth;
	mouseInfo.dy = dy / (int)gFontsHeight;
	dy -= mouseInfo.dy * (int)gFontsHeight;
	selection(p, oldstatus);
	oldstatus = mouseInfo.status;
}

#if defined (__NetBSD__) || defined (__OpenBSD__)
static void analyzeWSConsEvents(TVterm *p)
{
	static char oldstatus;
	static int dx, dy;
	struct wscons_event wscons_event;
	ssize_t nbytes;

	assert(initialized);
	assert(p != NULL);

	for (;;) {
		nbytes = read(mouse.fd, &wscons_event,
			      sizeof(struct wscons_event));
		if (nbytes != sizeof(struct wscons_event))
			break;
		switch (wscons_event.type) {
		case WSCONS_EVENT_MOUSE_UP:
			mouseInfo.status = 0;
			break;
		case WSCONS_EVENT_MOUSE_DOWN:
			if (wscons_event.value == 0)
				mouseInfo.status = MOUSE_BUTTON_LEFT;
			else if (wscons_event.value == 1)
				mouseInfo.status = MOUSE_BUTTON_MIDDLE;
			else if (wscons_event.value == 2)
				mouseInfo.status = MOUSE_BUTTON_RIGHT;
			else
				mouseInfo.status = 0;
			break;
		case WSCONS_EVENT_MOUSE_DELTA_X:
			dx += wscons_event.value;
			break;
		case WSCONS_EVENT_MOUSE_DELTA_Y:
			dy -= wscons_event.value;
			break;
		default:
			continue;
		}
		mouseInfo.dx = dx / (int)gFontsWidth;
		dx -= mouseInfo.dx * (int)gFontsWidth;
		mouseInfo.dy = dy / (int)gFontsHeight;
		dy -= mouseInfo.dy * (int)gFontsHeight;
		selection(p, oldstatus);
		oldstatus = mouseInfo.status;
	}
}
#endif

static void selection(TVterm *p, const char oldstatus)
{
	assert(initialized);
	assert(p != NULL);

	if (!mouse.enable)
		return;
	resetTimer();
	if (mouseInfo.dx || mouseInfo.dy) {
		mouseInfo.x += mouseInfo.dx;
		mouseInfo.y += mouseInfo.dy;
		if (mouseInfo.x < 0)
			mouseInfo.x = 0;
		else if (mouseInfo.x > p->xmax - 1)
			mouseInfo.x = p->xmax - 1;
		if (mouseInfo.y < 0)
			mouseInfo.y = 0;
		else if (mouseInfo.y > p->ymax - 1)
			mouseInfo.y = p->ymax - 1;
	}
	if (mouseInfo.status & MOUSE_BUTTON_LEFT) {
		if (!(oldstatus & MOUSE_BUTTON_LEFT)) {
			mouseInfo.sx = mouseInfo.x;
			mouseInfo.sy = mouseInfo.y;
		} else if (mouseInfo.dx || mouseInfo.dy) {
#ifdef ENABLE_SCROLLBACK
			/* TODO: history copy and paste */
			if (!keyboard_isScrollLocked() && !sb_is_on()) {
				vterm_reverseText(p, mouseInfo.sx, mouseInfo.sy,
						  mouseInfo.x, mouseInfo.y);
				vterm_refresh(p);
				vterm_reverseText(p, mouseInfo.sx, mouseInfo.sy,
						  mouseInfo.x, mouseInfo.y);
			}
#else
			if (!keyboard_isScrollLocked()) {
				vterm_reverseText(p, mouseInfo.sx, mouseInfo.sy,
						  mouseInfo.x, mouseInfo.y);
				vterm_refresh(p);
				vterm_reverseText(p, mouseInfo.sx, mouseInfo.sy,
						  mouseInfo.x, mouseInfo.y);
			}
#endif
		}
	} else if (oldstatus & MOUSE_BUTTON_LEFT) {
#ifdef ENABLE_SCROLLBACK
		/* TODO: history copy and paste */
		if (!keyboard_isScrollLocked() && !sb_is_on())
			vterm_copyText(p, mouseInfo.sx, mouseInfo.sy,
				       mouseInfo.x, mouseInfo.y);
#else
		if (!keyboard_isScrollLocked())
			vterm_copyText(p, mouseInfo.sx, mouseInfo.sy,
				       mouseInfo.x, mouseInfo.y);
		}
#endif
	}
	if (mouseInfo.status & mouse.pasteButton &&
	    !(oldstatus & mouse.pasteButton)) {
#ifdef ENABLE_SCROLLBACK
		/* TODO: history copy and paste */
		if (!keyboard_isScrollLocked() && !sb_is_on())
			vterm_pasteText(p);
#else
		if (!keyboard_isScrollLocked())
			vterm_pasteText(p);
#endif
	}
}

void mouse_getPackets(TVterm *p)
{
	static u_char packet[MAX_PACKET_SIZE];
	static int status;
	int i;
	ssize_t nbytes;
	u_char buf[BUF_SIZE];

	assert(initialized);
	assert(p != NULL);

#if defined (__NetBSD__) || defined (__OpenBSD__)
	if (mouse.type == MOUSE_TYPE_WSMOUSE) {
		analyzeWSConsEvents(p);
		return;
	}
#endif

	nbytes = read(mouse.fd, buf, BUF_SIZE);
	if (nbytes > 0) {
		for (i = 0; i < nbytes; i++) {
			if (status == 0) {
				if ((buf[i] & mouse.headMask) == mouse.headId) {
					packet[0] = buf[i];
					status = 1;
				}
				continue;
			}
			if (mouse.type != MOUSE_TYPE_PS2 &&
			    ((buf[i] & mouse.dataMask) || buf[i] == 0x80)) {
				status = 0;
				continue;
			}
			packet[status] = buf[i];
			status++;
			if (status == mouse.pkMax) {
				analyzePacket(p, packet);
				status = 0;
			}
		}
	}
}

