/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM/ME - J framebuffer terminal/Multilingual Enhancement -
 * Copyright (C) 2003 Fumitoshi UKAI <ukai@debian.or.jp>
 * Copyright (C) 1999 Noritoshi MASUICHI <nmasu@ma3.justnet.ne.jp>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <assert.h>
#include <err.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pcf.h"

#define TPCF_DEFAULT_FORMAT     (0)
#define TPCF_INKBOUNDS          (2)
#define TPCF_ACCEL_W_INKBOUNDS  (1)
#define TPCF_COMPRESSED_METRICS (1)

#define TPCF_PROPERTIES         (1 << 0)
#define TPCF_ACCELERATORS       (1 << 1)
#define TPCF_METRICS            (1 << 2)
#define TPCF_BITMAPS            (1 << 3)
#define TPCF_INK_METRICS        (1 << 4)
#define TPCF_BDF_ENCODINGS      (1 << 5)
#define TPCF_SWIDTHS            (1 << 6)
#define TPCF_GLYPH_NAMES        (1 << 7)
#define TPCF_BDF_ACCELERATORS   (1 << 8)

#define TPCF_INDEX_SIZE         (4 + 4 + 4 + 4)
#define TPCF_PROP_SIZE          (4 + 1 + 4)

static void die_file_eof(void);
static u_char read_u8(FILE *stream);
static u_int read_u16l(FILE *stream);
static u_int read_u16b(FILE *stream);
static u_int read_u16(FILE *stream, int e);
static int read_s16l(FILE *stream);
static int read_s16b(FILE *stream);
static int read_s16(FILE *stream, int e);
static u_int read_u32l(FILE *stream);
static u_int read_u32b(FILE *stream);
static u_int read_u32(FILE *stream, int e);
static int read_s32l(FILE *stream);
static int read_s32b(FILE *stream);
static int read_s32(FILE *stream, int e);
static void skip_nbytes(FILE *stream, size_t len);
static size_t seek_section(FILE *stream, size_t cur, size_t tag);
static void tpcfformat_load(TPcfFormat *p, FILE *stream);
static u_int to_u32l(char *p);
static void tpcfindex_load(TPcfIndex *p, FILE *stream);
#ifdef PCF_DEBUG
static void tpcfindex_debug(TPcfIndex *p);
#endif
static void tpcfprop_load(TPcfProp *p, FILE *stream, int e);
static void tpcfprops_init(TPcfProps *p);
static void tpcfprops_final(TPcfProps *p);
static size_t tpcfprops_load(TPcfProps *p, FILE *stream);
#ifdef PCF_DEBUG
static void tpcfprops_debug(TPcfProps *p);
#endif
static size_t tpcfmetric_load(TPcfMetric *p, FILE *stream, int e);
static size_t tpcfcmetric_load(TPcfMetric *p, FILE *stream);
#ifdef PCF_DEBUG
static void tpcfmetric_debug(TPcfMetric *p);
#endif
static void tpcfmetrics_init(TPcfMetrics *p);
static void tpcfmetrics_final(TPcfMetrics *p);
static size_t tpcfmetrics_load(TPcfMetrics *p, FILE *stream);
static size_t tpcfaccel_load(TPcfAccel *p, FILE *stream);
#ifdef PCF_DEBUG
static void tpcfaccel_debug(TPcfAccel *p);
#endif
static void tpcfbitmap_init(TPcfBitmap *p);
static void tpcfbitmap_final(TPcfBitmap *p);
static void tpcfbitmap_swap(TPcfBitmap *p, int e, int aline);
static size_t tpcfbitmap_load(TPcfBitmap *p, FILE *stream);
static void tpcfencode_init(TPcfEncode *p);
static void tpcfencode_final(TPcfEncode *p);
static size_t tpcfencode_load(TPcfEncode *p, FILE *stream);
#ifdef PCF_DEBUG
static void tpcfencode_debug(TPcfEncode *p);
#endif
static TPcfIndex *tpcf_search_section(TPcf *p, u_int type);
#ifdef PCF_DEBUG
static const char *tpcf_get_string_prop(TPcf *p, const char *key);
#endif

static void die_file_eof(void)
{
	errx(1, "Unexpected EOF");
}

static u_char read_u8(FILE *stream)
{
	u_char b;

	assert(stream != NULL);

	if (fread(&b, 1, 1, stream) != 1) die_file_eof();
	return b;
}

/* little endian */
static u_int read_u16l(FILE *stream)
{
	u_char b[2];

	assert(stream != NULL);

	if (fread(b, 2, 1, stream) != 1) die_file_eof();
	return b[0] + (b[1] << 8);
}

/* big endian */
static u_int read_u16b(FILE *stream)
{
	u_char b[2];

	assert(stream != NULL);

	if (fread(b, 2, 1, stream) != 1) die_file_eof();
	return b[1] + (b[0] << 8);
}

static u_int read_u16(FILE *stream, int e)
{
	assert(stream != NULL);

	return (e & 1) ? read_u16b(stream) : read_u16l(stream);
}

/* little endian */
static int read_s16l(FILE *stream)
{
	u_int t;

	assert(stream != NULL);

	t = read_u16l(stream);
	if (t >= 0x8000) t |= ~0xffff;
	return (int)t;
}

/* big endian */
static int read_s16b(FILE *stream)
{
	u_int t;

	assert(stream != NULL);

	t = read_u16b(stream);
	if (t >= 0x8000) t |= ~0xffff;
	return (int)t;
}

static int read_s16(FILE *stream, int e)
{
	assert(stream != NULL);

	return (e & 1) ? read_s16b(stream) : read_s16l(stream);
}

/* little endian */
static u_int read_u32l(FILE *stream)
{
	u_char b[4];

	assert(stream != NULL);

	if (fread(b, 4, 1, stream) != 1) die_file_eof();
	return b[0] + (b[1] << 8) + (b[2] << 16) + (b[3] << 24);
}

/* big endian */
static u_int read_u32b(FILE *stream)
{
	u_char b[4];

	assert(stream != NULL);

	if (fread(b, 4, 1, stream) != 1) die_file_eof();
	return b[3] + (b[2] << 8) + (b[1] << 16) + (b[0] << 24);
}

static u_int read_u32(FILE *stream, int e)
{
	assert(stream != NULL);

	return (e & 1) ? read_u32b(stream) : read_u32l(stream);
}

/* little endian */
static int read_s32l(FILE *stream)
{
	u_char b[4];

	assert(stream != NULL);

	if (fread(b, 4, 1, stream) != 1) die_file_eof();
	return b[0] + (b[1] << 8) + (b[2] << 16) + (b[3] << 24);
}

/* big endian */
static int read_s32b(FILE *stream)
{
	u_char b[4];

	assert(stream != NULL);

	if (fread(b, 4, 1, stream) != 1) die_file_eof();
	return b[3] + (b[2] << 8) + (b[1] << 16) + (b[0] << 24);
}

static int read_s32(FILE *stream, int e)
{
	assert(stream != NULL);

	return (e & 1) ? read_s32b(stream) : read_s32l(stream);
}

static void skip_nbytes(FILE *stream, size_t nbytes)
{
	char dummy[1024];
	size_t n;

	assert(stream != NULL);

	while (nbytes != 0) {
		n = nbytes > sizeof(dummy) ? sizeof(dummy) : nbytes;
		if (fread(dummy, n, 1, stream) != 1) die_file_eof();
		nbytes -= n;
	}
}

static size_t seek_section(FILE *stream, size_t cur, size_t tag)
{
	size_t n;
	size_t d;
	char dummy[1024];

	assert(stream != NULL);

	if (cur > tag)
		errx(1, "(FONT): backward seeking.");
	else if (cur == tag)
		return tag;
	n = tag - cur;
	while (n != 0) {
		d = n > 1024 ? 1024 : n;
		if (fread(dummy, d, 1, stream) != 1) die_file_eof();
		n -= d;
	}
	return tag;
}

static void tpcfformat_load(TPcfFormat *p, FILE *stream)
{
	u_int t;

	assert(p != NULL);
	assert(stream != NULL);

	t = read_u32l(stream);
#if PCF_DEBUG
	fprintf(stderr, "[[FORMAT: %x = ", t);
#endif
	p->id = (t >> 8) & 0xfff;
	p->scan = 1 << ((t >> 4) & 0x3);
	p->obit = (t >> 3) & 1;
	p->obyte = (t >> 2) & 1;
	p->glyphaline = t & 0x3;
#if PCF_DEBUG
	switch (p->id) {
	case TPCF_DEFAULT_FORMAT:
		fprintf(stderr, "default ");
		break;
	case TPCF_INKBOUNDS:
		fprintf(stderr, "inkbounds ");
		break;
	case TPCF_ACCEL_W_INKBOUNDS:
		fprintf(stderr, "accel/ink|compr/metric ");
		break;
	default:
		fprintf(stderr, "???");
		break;
	}
	fprintf(stderr, "scan:%d bytes ", p->scan);
	if (p->obit)
		fprintf(stderr, "MSB ");
	else
		fprintf(stderr, "LSB ");
	if (p->obyte)
		fprintf(stderr, "BE ");
	else
		fprintf(stderr, "LE ");
	fprintf(stderr, "g:%d (%d bytes)", p->glyphaline, 1 << p->glyphaline);
	fprintf(stderr, "]]\n");
#endif
}

static u_int to_u32l(char *p)
{
	u_int i, j;

	assert(p != NULL);

	j = 0;
	for (i = 0; i < 4; i++)
		j += ((u_char)p[i]) << (i * 8);
	return j;
}

static void tpcfindex_load(TPcfIndex *p, FILE *stream)
{
	assert(p != NULL);
	assert(stream != NULL);

	p->type   = read_u32l(stream);
	tpcfformat_load(&(p->fmt), stream);
	p->size   = read_s32l(stream);
	p->offset = read_s32l(stream);
}

#ifdef PCF_DEBUG
static void tpcfindex_debug(TPcfIndex *p)
{
	assert(p != NULL);

	fprintf(stderr, "type=%u:", p->type);
	switch (p->type) {
	case TPCF_PROPERTIES:
		fprintf(stderr, "prop ");
		break;
	case TPCF_ACCELERATORS:
		fprintf(stderr, "accel ");
		break;
	case TPCF_METRICS:
		fprintf(stderr, "metric ");
		break;
	case TPCF_BITMAPS:
		fprintf(stderr, "bitmaps ");
		break;
	case TPCF_INK_METRICS:
		fprintf(stderr, "inkmetric ");
		break;
	case TPCF_BDF_ENCODINGS:
		fprintf(stderr, "bdfencodings ");
		break;
	case TPCF_SWIDTHS:
		fprintf(stderr, "swidth ");
		break;
	case TPCF_GLYPH_NAMES:
		fprintf(stderr, "glyphnames ");
		break;
	case TPCF_BDF_ACCELERATORS:
		fprintf(stderr, "bdfaccel ");
		break;
	default:
		fprintf(stderr, "???");
		break;
	}
	fprintf(stderr, ", size=%u, offset=%u", p->size, p->offset);
}
#endif

static void tpcfprop_load(TPcfProp *p, FILE *stream, int e)
{
	assert(p != NULL);
	assert(stream != NULL);

	p->name = read_s32(stream, e);
	p->strq = read_u8(stream);
	p->val  = read_s32(stream, e);
}

static void tpcfprops_init(TPcfProps *p)
{
	assert(p != NULL);

	p->nProp   = 0;
	p->props   = NULL;
	p->nSlen   = 0;
	p->strings = NULL;
}

static void tpcfprops_final(TPcfProps *p)
{
	assert(p != NULL);

	p->nProp = 0;
	if (p->props != NULL) {
		free(p->props);
		p->props = NULL;
	}
	p->nSlen = 0;
	if (p->strings != NULL) {
		free(p->strings);
		p->strings = NULL;
	}
}

static size_t tpcfprops_load(TPcfProps *p, FILE *stream)
{
	int i;
	int dummy;
	TPcfFormat fmt;

	assert(p != NULL);
	assert(stream != NULL);

	tpcfformat_load(&fmt, stream);
	p->e = fmt.obit;
	p->nProp = read_s32(stream, p->e);
#if 0
	fprintf(stderr, "Prop = %d\n", p->nProp);
#endif
	p->props = (TPcfProp *)malloc(sizeof(TPcfProp) * p->nProp);
	if (p->props == NULL)
		err(1, "malloc()");
	for (i = 0; i < p->nProp; i++)
		tpcfprop_load(&(p->props[i]), stream, p->e);
	dummy = 3 - (((TPCF_PROP_SIZE * p->nProp) + 3) % 4);
	for (i = 0; i < dummy; i++)
		read_u8(stream);
	p->nSlen = read_s32(stream, p->e);
	p->strings = (char *)malloc(sizeof(u_char) * p->nSlen);
	if (p->strings == NULL)
		err(1, "malloc()");
	if (fread(p->strings, p->nSlen, 1, stream) != 1) die_file_eof();
	return 8 + p->nProp * TPCF_PROP_SIZE + dummy + 4 + p->nSlen;
}

#ifdef PCF_DEBUG
static void tpcfprops_debug(TPcfProps *p)
{
	int i;

	assert(p != NULL);

	for (i = 0; i < p->nProp; i++) {
		if (p->props[i].strq) {
			fprintf(stderr, "%s = \"%s\"\n",
			       p->strings + (p->props[i].name),
			       p->strings + (p->props[i].val));
		} else {
			fprintf(stderr, "%s = %d\n",
			       p->strings + (p->props[i].name),
			       p->props[i].val);
		}
	}
}
#endif

static size_t tpcfmetric_load(TPcfMetric *p, FILE *stream, int e)
{
	assert(p != NULL);
	assert(stream != NULL);

	p->leftsb  = read_s16(stream, e);
	p->rightsb = read_s16(stream, e);
	p->width   = read_s16(stream, e);
	p->ascent  = read_s16(stream, e);
	p->descent = read_s16(stream, e);
	p->attr    = read_s16(stream, e);
	return 2 * 6;
}

static size_t tpcfcmetric_load(TPcfMetric *p, FILE *stream)
{
	assert(p != NULL);
	assert(stream != NULL);

	p->leftsb  = read_u8(stream) - 0x80;
	p->rightsb = read_u8(stream) - 0x80;
	p->width   = read_u8(stream) - 0x80;
	p->ascent  = read_u8(stream) - 0x80;
	p->descent = read_u8(stream) - 0x80;
	p->attr    = 0;
	return 1 * 5;
}

#ifdef PCF_DEBUG
static void tpcfmetric_debug(TPcfMetric *p)
{
	assert(p != NULL);

	fprintf(stderr, "[METRIC:%d<>%d(%d):%d^v%d:%d]",
	       p->leftsb, p->rightsb, p->width,
	       p->ascent, p->descent, p->attr);
}
#endif

static void tpcfmetrics_init(TPcfMetrics *p)
{
	assert(p != NULL);

	p->nMetrics = 0;
	p->metric   = NULL;
}

static void tpcfmetrics_final(TPcfMetrics *p)
{
	assert(p != NULL);

	p->nMetrics = 0;
	if (p->metric != NULL) {
		free(p->metric);
		p->metric = NULL;
	}
}

static size_t tpcfmetrics_load(TPcfMetrics *p, FILE *stream)
{
	TPcfFormat fmt;
	int i;
	size_t r;

	assert(p != NULL);
	assert(stream != NULL);

	r = 0;
	tpcfformat_load(&fmt, stream);
	if (fmt.id == TPCF_DEFAULT_FORMAT) {
		p->nMetrics = read_s32(stream, fmt.obit);
#ifdef PCF_DEBUG
		fprintf(stderr, "nmetric(default): %d\n", p->nMetrics);
#endif
		p->metric = (TPcfMetric *)malloc(sizeof(TPcfMetric) * p->nMetrics);
		if (p->metric == NULL)
			err(1, "malloc()");
		for (i = 0; i < p->nMetrics; i++)
			r += tpcfmetric_load(&(p->metric[i]), stream, fmt.obit);
		return 8 + r;
	} else if (fmt.id == TPCF_COMPRESSED_METRICS) {
		p->nMetrics = read_u16(stream, fmt.obit);
#ifdef PCF_DEBUG
		fprintf(stderr, "nmetric(compressed): %d\n", p->nMetrics);
#endif
		p->metric = (TPcfMetric *)malloc(sizeof(TPcfMetric) * p->nMetrics);
		if (p->metric == NULL)
			err(1, "malloc()");
		for (i = 0; i < p->nMetrics; i++)
			r += tpcfcmetric_load(&(p->metric[i]), stream);
		return 6 + r;
	}
	return 4;
}

static size_t tpcfaccel_load(TPcfAccel *p, FILE *stream)
{
	TPcfFormat fmt;
	size_t r;

	assert(p != NULL);
	assert(stream != NULL);

	tpcfformat_load(&fmt, stream);

	/* bool8: noOverlap */
	/* bool8: constantMetrics */
	skip_nbytes(stream, 2);

	/* bool8: terminalFont */
	p->termf = read_u8(stream);
#if 0 /* XXX: unifont is not terminal font */
	if (!p->termf)
		errx(1, "(FONT) : Not terminal font %02x.", p->termf);

#endif
	/* bool8: contantWidth */
	/* bool8: inkInside */
	/* bool8: inkMetrics */
	/* bool8: drawDirection */
	/* bool8: dummy (padding) */
	/* int32: fontAscent */
	/* int32: fontDescent */
	/* int32: maxOverlap */
	skip_nbytes(stream, 5 + 4 * 3);

	/* metric: minBounds */
	r = tpcfmetric_load(&(p->metric), stream, fmt.obit);

	/* metric: maxBounds */

	/* if format.id == PCF_ACCEL_W_INKBOUNDS */
	/*   metric: ink_minBounds */
	/*   metric: ink_maxBounds */
	/* endif */
	return 24 + r;
}

#ifdef PCF_DEBUG
static void tpcfaccel_debug(TPcfAccel *p)
{
	assert(p != NULL);

	fprintf(stderr, "[ACCEL:");
	tpcfmetric_debug(&(p->metric));
	fprintf(stderr, "]\n");
}
#endif

static void tpcfbitmap_init(TPcfBitmap *p)
{
	assert(p != NULL);

	p->maps    = 0;
	p->offsets = NULL;
	p->bitmap  = NULL;
}

static void tpcfbitmap_final(TPcfBitmap *p)
{
	assert(p != NULL);

	p->maps = 0;
	if (p->offsets != NULL) {
		free(p->offsets);
		p->offsets = NULL;
	}
	if (p->bitmap != NULL) {
		free(p->bitmap);
		p->bitmap = NULL;
	}
}

static void tpcfbitmap_swap(TPcfBitmap *p, int e, int aline)
{
	u_int len;
	u_int i;
	u_char *bp;
	u_char c;
	u_char d;

	assert(p != NULL);

	len = p->mapsize[aline];
	bp = p->bitmap;
	if (!e) {       /* LSB <==> MSB */
		for (i = 0; i < len; i++) {
			c = *bp;
			d  = (c & 0x01) ? 0x80 : 0x00;
			d |= (c & 0x02) ? 0x40 : 0x00;
			d |= (c & 0x04) ? 0x20 : 0x00;
			d |= (c & 0x08) ? 0x10 : 0x00;
			d |= (c & 0x10) ? 0x08 : 0x00;
			d |= (c & 0x20) ? 0x04 : 0x00;
			d |= (c & 0x40) ? 0x02 : 0x00;
			d |= (c & 0x80) ? 0x01 : 0x00;
			*bp++ = d;
		}
	}

}

static size_t tpcfbitmap_load(TPcfBitmap *p, FILE *stream)
{
	TPcfFormat fmt;
	size_t r;
	int i;

	assert(p != NULL);
	assert(stream != NULL);

	tpcfformat_load(&fmt, stream);
	p->aline = 1 << fmt.glyphaline;
	p->galine = fmt.glyphaline;

	p->maps = read_s32(stream, fmt.obit);
	p->offsets = (u_int *)malloc(sizeof(u_int) * p->maps);
	if (p->offsets == NULL)
		err(1, "malloc()");
	for (i = 0; i < p->maps; i++)
		p->offsets[i] = read_u32(stream, fmt.obit);
	for (i = 0; i < 4; i++)
		p->mapsize[i] = read_u32(stream, fmt.obit);
	r = p->mapsize[fmt.glyphaline];
#if PCF_DEBUG
	fprintf(stderr, "BITMAP SIZE : %d bytes - %d (%d bytes/line)\n",
		r, fmt.glyphaline, p->aline);
#endif
	p->bitmap = (u_char *)malloc(r);
	if (p->bitmap == NULL)
		err(1, "malloc()");
	if (fread(p->bitmap, r, 1, stream) != 1) die_file_eof();

	tpcfbitmap_swap(p, fmt.obit, fmt.glyphaline);

	return 8 + p->maps * 4 + 4 * 4 + r;
}

static void tpcfencode_init(TPcfEncode *p)
{
	assert(p != NULL);

	p->table = NULL;
}

static void tpcfencode_final(TPcfEncode *p)
{
	assert(p != NULL);

	if (p->table != NULL) {
		free(p->table);
		p->table = NULL;
	}
}

static size_t tpcfencode_load(TPcfEncode *p, FILE *stream)
{
	TPcfFormat fmt;
	size_t r;
	int i;

	assert(p != NULL);
	assert(stream != NULL);

	tpcfformat_load(&fmt, stream);

	p->colf = read_s16(stream, fmt.obit);
	p->coll = read_s16(stream, fmt.obit);
	p->rowf = read_s16(stream, fmt.obit);
	p->rowl = read_s16(stream, fmt.obit);
	p->defa = read_s16(stream, fmt.obit);

	r = (p->coll - p->colf + 1) * (p->rowl - p->rowf + 1);
	p->table = (int *)malloc(sizeof(int) * r);
	if (p->table == NULL)
		err(1, "malloc()");
	for (i = 0; i < r; i++)
		p->table[i] = read_u16(stream, fmt.obit);

	return 14 + 2 * r;
}

#ifdef PCF_DEBUG
static void tpcfencode_debug(TPcfEncode *p)
{
	assert(p != NULL);

	fprintf(stderr, "ENCODE[");
	fprintf(stderr, "col:%d-%d[%x-%x]", p->colf, p->coll, p->colf, p->coll);
	fprintf(stderr, "row:%d-%d[%x-%x]", p->rowf, p->rowl, p->rowf, p->rowl);
	fprintf(stderr, "defch: %d ", p->defa);
	fprintf(stderr, "n = %d]\n", (p->coll - p->colf + 1) * (p->rowl - p->rowf + 1));
}
#endif

void pcf_initialize(TPcf *p)
{
	assert(p != NULL);

	p->nIdx = 0;
	p->idxs = NULL;
	tpcfprops_init(&(p->props));
	tpcfmetrics_init(&(p->metrics));
	tpcfbitmap_init(&(p->bitmap));
	tpcfencode_init(&(p->encode));
}

void pcf_finalize(TPcf *p)
{
	assert(p != NULL);

	p->nIdx = 0;
	if (p->idxs != NULL) {
		free(p->idxs);
		p->idxs = NULL;
	}
	tpcfprops_final(&(p->props));
	tpcfmetrics_final(&(p->metrics));
	tpcfbitmap_final(&(p->bitmap));
	tpcfencode_final(&(p->encode));
}

static TPcfIndex *tpcf_search_section(TPcf *p, u_int type)
{
	int i;

	assert(p != NULL);

	for (i = 0; i < p->nIdx; i++)
		if (type == p->idxs[i].type)
			return &(p->idxs[i]);
	return NULL;
}

void tpcf_load(TPcf *p, FILE *stream)
{
	char magic[] = { 1, 'f', 'c', 'p' };
	/* --- */
	size_t readp = 0;
	TPcfIndex *q;
	TPcfIndex *qt;
	int i;

	assert(p != NULL);
	assert(stream != NULL);

	/* read header and prepare index-area */
	if (read_u32l(stream) != to_u32l(magic))
		errx(1, "(FONT): PCF file format error: Bad signature.");
	p->nIdx = read_u32l(stream);
	p->idxs = (TPcfIndex *)malloc(sizeof(TPcfIndex) * p->nIdx);
	if (p->idxs == NULL)
		err(1, "malloc()");
	readp += 8;

	/* read index */
	for (i = 0; i < p->nIdx; i++) {
		tpcfindex_load(&(p->idxs[i]), stream);
#if PCF_DEBUG
		fprintf(stderr, "SECT %d :", i);
		tpcfindex_debug(&(p->idxs[i]));
		fprintf(stderr, "\n");
#endif
		readp += TPCF_INDEX_SIZE;
	}
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d\n", readp);
#endif

	/* read properties */
	q = tpcf_search_section(p, TPCF_PROPERTIES);
	if (!q)
		errx(1, "(FONT): Properties section not exist.");

	readp = seek_section(stream, readp, q->offset);
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d\n", readp);
	fprintf(stderr, "load PROPS\n");
#endif
	readp += tpcfprops_load(&(p->props), stream);
#if PCF_DEBUG
	tpcfprops_debug(&(p->props));
	fprintf(stderr, "total gain = %d\n", readp);
#endif

	/* read accelerator */
	q = tpcf_search_section(p, TPCF_ACCELERATORS);
	qt = tpcf_search_section(p, TPCF_BDF_ACCELERATORS);
	if (!qt) {
		if (!q)
			errx(1, "(FONT): Accelerators sections not exist.");
#if PCF_DEBUG
		fprintf(stderr, "total gain = %d:%ld\n", readp, ftell(stream));
#endif
		readp = seek_section(stream, readp, q->offset);
#if PCF_DEBUG
		fprintf(stderr, "total gain = %d:%ld\n", readp, ftell(stream));
		fprintf(stderr, "load ACCEL\n");
#endif
		readp += tpcfaccel_load(&(p->accel), stream);
#if PCF_DEBUG
		tpcfaccel_debug(&(p->accel));
		fprintf(stderr, "total gain = %d\n", readp);
#endif
	}
	/* read metric: for multiwidth font (unifont) ? */
	q = tpcf_search_section(p, TPCF_METRICS);
	if (!q)
		errx(1, "(FONT): Metric section not exist.");
	readp = seek_section(stream, readp, q->offset);
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d\n", readp);
	fprintf(stderr, "load METRIC\n");
#endif
	readp += tpcfmetrics_load(&(p->metrics), stream);
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d\n", readp);
#endif

	/* read bitmap */
	q = tpcf_search_section(p, TPCF_BITMAPS);
	if (!q)
		errx(1, "(FONT): Bitmap section not exist.");
	readp = seek_section(stream, readp, q->offset);
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d\n", readp);
	fprintf(stderr, "load BITMAP\n");
#endif
	readp += tpcfbitmap_load(&(p->bitmap), stream);
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d\n", readp);
#endif

	/* read encoding */
	q = tpcf_search_section(p, TPCF_BDF_ENCODINGS);
	if (!q)
		errx(1, "(FONT): Encoding section not exist.");
	readp = seek_section(stream, readp, q->offset);
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d:%ld\n", readp, ftell(stream));
	fprintf(stderr, "load ENCODE\n");
#endif
	readp += tpcfencode_load(&(p->encode), stream);
#if PCF_DEBUG
	tpcfencode_debug(&(p->encode));
	fprintf(stderr, "total gain = %d:%ld\n", readp, ftell(stream));
#endif

	/* read bdf-accelerator */
	q = tpcf_search_section(p, TPCF_BDF_ACCELERATORS);
	if (!q)
		errx(1, "(FONT): Bdf_accelerators section not exist.");
	readp = seek_section(stream, readp, q->offset);
#if PCF_DEBUG
	fprintf(stderr, "total gain = %d\n", readp);
	fprintf(stderr, "load BDF_ACCEL\n");
#endif
	readp += tpcfaccel_load(&(p->accel), stream);
#if PCF_DEBUG
	tpcfaccel_debug(&(p->accel));
	fprintf(stderr, "total gain = %d\n", readp);
#endif
}

void tpcf_as_tfont(TPcf *p, TFont *q)
{
	int offset;
	int gs;
	int bs;
	int i;
	int ii;

	assert(p != NULL);
	assert(q != NULL);

	q->width = p->accel.metric.rightsb - p->accel.metric.leftsb;
	q->height = p->accel.metric.ascent + p->accel.metric.descent;

	bs = p->bitmap.mapsize[p->bitmap.galine];
	q->bitmap = (u_char *)malloc(bs);
	if (q->bitmap == NULL)
		err(1, "malloc()");
	memcpy(q->bitmap, p->bitmap.bitmap, bs);

	q->bytew = p->bitmap.aline;
	q->bytec = q->bytew * q->height;
	q->colf = p->encode.colf;
	q->coll = p->encode.coll;
	q->rowf = p->encode.rowf;
	q->rowl = p->encode.rowl;
	q->colspan = q->coll - q->colf + 1;

	q->dglyph = (u_char *)malloc(q->bytec);
	if (q->dglyph == NULL)
		err(1, "malloc()");
	memset(q->dglyph, 0xff, q->bytec);

	gs = (q->coll - q->colf + 1) * (q->rowl - q->rowf + 1);
	q->glyph = (u_char **)malloc(sizeof(u_char *) * gs);
	if (q->glyph == NULL)
		err(1, "malloc()");
	q->glyph_width = NULL;
	if (p->accel.termf == 0) {
		/* not terminal font */
		q->glyph_width = (u_int *)malloc(sizeof(u_int) * gs);
		if (q->glyph_width == NULL)
			err(1, "malloc()");
	}
	for (i = 0; i < gs; i++) {
		ii = p->encode.table[i];
		if (ii == 0xffff) {
			q->glyph[i] = q->dglyph;
			if (q->glyph_width != NULL)
				q->glyph_width[i] = q->width; /* XXX */
		} else {
			offset = p->bitmap.offsets[ii];
			q->glyph[i] = q->bitmap + offset;
			if (q->glyph_width != NULL) {
				TPcfMetric *m;
				m = &p->metrics.metric[ii];
				q->glyph_width[i] = m->rightsb - m->leftsb;
			}
		}
	}
}

#if PCF_DEBUG
static const char *tpcf_get_string_prop(TPcf *p, const char *key)
{
	int i;
	const char *strb;

	assert(p != NULL);
	assert(key != NULL);

	strb = p->props.strings;
	for (i = 0; i < p->props.nProp; i++) {
		if (p->props.props[i].strq &&
		    strcmp(strb + (p->props.props[i].name), key) == 0)
			return strb + (p->props.props[i].val);
	}
	return NULL;
}
#endif

#if PCF_DEBUG
#include "main.h"

/* cc -I. -o p -DPCF_DEBUG=1 pcf.c */

void dump_bits(int ch)
{
	int b;

	/* fprintf(stderr, "%02x:", ch); */
	for (b = 7; b >= 0; --b) {
		if ((ch & (1 << b)) != 0)
			fprintf(stderr, "*");
		else
			fprintf(stderr, "-");
	}
	fprintf(stderr, " ");
}

int main(int argc, char *argv[])
{
	FILE *stream;
	char buf[512];
	TPcf tpcf;

	if (argc < 2) {
		fprintf(stderr, stderr, "usage: %s <fontfile> [idx]\n", argv[0]);
		return EXIT_FAILURE;
	}
	snprintf(buf, 512, "gunzip < %s", argv[1]);
	stream = popen(buf, "r");

	tpcf_load(&tpcf, stream);
	if (argc > 2) {
		int idx;
		int eidx;
		int offset;
		char *p;
		int i;
		int height, width;
		idx = strtol(argv[2], NULL, 0);
		fprintf(stderr, "idx:%d (0x%04x) => ", idx, idx);
		fprintf(stderr, "colspan=%d ",
		       (tpcf.encode.coll - tpcf.encode.colf + 1));
		eidx = ((idx >> 8) - tpcf.encode.rowf) *
		       (tpcf.encode.coll - tpcf.encode.colf + 1)
		       + (idx & 0x0ff) - tpcf.encode.colf;
		fprintf(stderr, "=> %d (0x%04x)\n", eidx, eidx);
		offset = tpcf.encode.table[eidx];
		fprintf(stderr, "encode.offset:%d (0x%04x)\n", offset, offset);
		if (offset == 0xffff)
			fprintf(stderr, "no glyph\n");
		else {
			TPcfMetric *m;
			int mheight, mwidth;
			if (offset >= tpcf.metrics.nMetrics) {
				fprintf(stderr, "out of error? metric: %d > %d\n",
				       offset, tpcf.metrics.nMetrics);
				return EXIT_FAILURE;
			}
			m = &tpcf.metrics.metric[offset];
			offset = tpcf.bitmap.offsets[offset];
			fprintf(stderr, "bitmap.offset:%d\n", offset);
			p = &tpcf.bitmap.bitmap[offset];
			height = tpcf.accel.metric.ascent + tpcf.accel.metric.descent;
			width = tpcf.accel.metric.rightsb - tpcf.accel.metric.leftsb;
			mheight = m->ascent + m->descent;
			mwidth = m->rightsb - m->leftsb;
			fprintf(stderr, "(h:%d * w:%d) h:%d * w:%d \n",
			       height, width, mheight, mwidth);
			for (i = 0; i < height; i++) {
				int j;
				for (j = 0; j < tpcf.bitmap.aline; j++) {
					dump_bits(*p & 0x0ff); p++;
				}
				fprintf(stderr, "\n");
			}
		}
	}
	return EXIT_SUCCESS;
}
#endif

