/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM/ME - J framebuffer terminal/Multilingual Enhancement -
 * Copyright (C) 2003 Fumitoshi UKAI <ukai@debian.or.jp>
 * Copyright (C) 1999 Noritoshi MASUICHI <nmasu@ma3.justnet.ne.jp>
 *
 * KON2 - Kanji ON Console -
 * Copyright (C) 1992,1993 Atusi MAEDA <mad@math.keio.ac.jp>
 * Copyright (C) 1992-1996 Takashi MANABE <manabe@papilio.tutics.tut.ac.jp>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <assert.h>
#include <stdlib.h>

#include "palette.h"
#include "pen.h"

void pen_initialize(TPen *p)
{
	assert(p != NULL);

	p->prev = NULL;
	p->x = 0;
	p->y = 0;
	p->fcol = 7;
	p->bcol = 0;
	p->attr = 0;
}

void pen_finalize(TPen *p)
{
	TPen *q;

	assert(p != NULL);

	q = p->prev;
	p->prev = NULL;
	if (q != NULL) {
		pen_finalize(q);
		free(q);
	}
}

void pen_copy(TPen *p, TPen *q)
{
	assert(p != NULL);

	p->prev = NULL;
	p->x = q->x;
	p->y = q->y;
	p->fcol = q->fcol;
	p->bcol = q->bcol;
	p->attr = q->attr;
}

inline void pen_resetAttribute(TPen *p)
{
	assert(p != NULL);

	p->fcol = 7;
	p->bcol = 0;
	p->attr = 0;
}

void pen_setHighlight(TPen *p, const bool status)
{
	assert(p != NULL);

	if (status) {
		p->attr |= PEN_ATTR_HIGHLIGHT;
#ifdef ENABLE_256_COLOR
		if (p->fcol < 16) p->fcol |= 8;
#else
		if (p->fcol != 0) p->fcol |= 8;
#endif
	} else {
		p->attr &= ~PEN_ATTR_HIGHLIGHT;
#ifdef ENABLE_256_COLOR
		if (p->fcol < 16) p->fcol &= ~8;
#else
		p->fcol &= ~8;
#endif
	}
}

void pen_setUnderline(TPen *p, const bool status)
{
	assert(p != NULL);

	if (status) {
		p->attr |= PEN_ATTR_UNDERLINE;
#ifdef ENABLE_256_COLOR
		if (p->bcol < 16) p->bcol |= 8;
#else
		p->bcol |= 8;
#endif
	} else {
		p->attr &= ~PEN_ATTR_UNDERLINE;
#ifdef ENABLE_256_COLOR
		if (p->bcol < 16) p->bcol &= ~8;
#else
		p->bcol &= ~8;
#endif
	}
}

void pen_setReverse(TPen *p, const bool status)
{
	u_char swp;

	assert(p != NULL);

	if (status) {
		if (!(p->attr & PEN_ATTR_REVERSE)) {
			p->attr |= PEN_ATTR_REVERSE;
#ifdef ENABLE_256_COLOR
			if (p->fcol < 16) {
				swp = p->fcol & ~8;
				if (p->attr & PEN_ATTR_UNDERLINE)
					swp |= 8;
			} else
				swp = p->fcol;
			if (p->bcol < 16) {
				p->fcol = p->bcol & ~8;
				if (p->attr & PEN_ATTR_HIGHLIGHT && p->fcol)
					p->fcol |= 8;
			} else
				p->fcol = p->bcol;
			p->bcol = swp;
#else
			swp = p->fcol & ~8;
			if (p->attr & PEN_ATTR_UNDERLINE)
				swp |= 8;
			p->fcol = p->bcol & ~8;
			if (p->attr & PEN_ATTR_HIGHLIGHT && p->fcol)
				p->fcol |= 8;
			p->bcol = swp;
#endif
		}
	} else {
		if (p->attr & PEN_ATTR_REVERSE) {
			p->attr &= ~PEN_ATTR_REVERSE;
#ifdef ENABLE_256_COLOR
			if (p->fcol < 16) {
				swp = p->fcol & ~8;
				if (p->attr & PEN_ATTR_UNDERLINE)
					swp |= 8;
			} else
				swp = p->fcol;
			if (p->bcol < 16) {
				p->fcol = p->bcol & ~8;
				if (p->attr & PEN_ATTR_HIGHLIGHT && p->fcol)
					p->fcol |= 8;
			} else
				p->fcol = p->bcol;
			p->bcol = swp;
#else
			swp = p->fcol & ~8;
			if (p->attr & PEN_ATTR_UNDERLINE)
				swp |= 8;
			p->fcol = p->bcol & ~8;
			if (p->attr & PEN_ATTR_HIGHLIGHT && p->fcol)
				p->fcol |= 8;
			p->bcol = swp;
#endif
		}
	}
}

void pen_setColor(TPen *p, const int color)
{
	u_char t;

	assert(p != NULL);

	if (color >= 30 && color <= 37) {
		t = palette_ansiToVGA(color - 30);
		if (p->attr & PEN_ATTR_REVERSE) {
			if (p->attr & PEN_ATTR_UNDERLINE) t |= 8;
			p->bcol = t;
		} else {
			if (p->attr & PEN_ATTR_HIGHLIGHT) t |= 8;
			p->fcol = t;
		}
	} else if (color == 39) {
		t = 7;
		if (p->attr & PEN_ATTR_REVERSE) {
			if (p->attr & PEN_ATTR_UNDERLINE) t |= 8;
			p->bcol = t;
		} else {
			if (p->attr & PEN_ATTR_HIGHLIGHT) t |= 8;
			p->fcol = t;
		}
	} else if (color >= 40 && color <= 47) {
		t = palette_ansiToVGA(color - 40);
		if (p->attr & PEN_ATTR_REVERSE) {
			if (p->attr & PEN_ATTR_HIGHLIGHT) t |= 8;
			p->fcol = t;
		} else {
			if (p->attr & PEN_ATTR_UNDERLINE) t |= 8;
			p->bcol = t;
		}
	} else if (color == 49) {
		t = 0;
		if (p->attr & PEN_ATTR_REVERSE) {
			if (p->attr & PEN_ATTR_HIGHLIGHT) t |= 8;
			p->fcol = t;
		} else {
			if (p->attr & PEN_ATTR_UNDERLINE) t |= 8;
			p->bcol = t;
		}
	}
}

#ifdef ENABLE_256_COLOR
void pen_set256Color(TPen *p, const int mode, const int color)
{
	u_char t;

	assert(p != NULL);

	if (color < 0 || color > 255)
		return;
	t = palette_ansiToVGA(color);
	if (mode == 38) {
		if (p->attr & PEN_ATTR_REVERSE) {
			if (t < 16 && p->attr & PEN_ATTR_UNDERLINE) t |= 8;
			p->bcol = t;
		} else {
			if (t < 16 && p->attr & PEN_ATTR_HIGHLIGHT) t |= 8;
			p->fcol = t;
		}
	} else if (mode == 48) {
		if (p->attr & PEN_ATTR_REVERSE) {
			if (t < 16 && p->attr & PEN_ATTR_HIGHLIGHT) t |= 8;
			p->fcol = t;
		} else {
			if (t < 16 && p->attr & PEN_ATTR_UNDERLINE) t |= 8;
			p->bcol = t;
		}
	}
}
#endif

