/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM Unofficial eXtension Patch
 * Copyright (C) 2006 Young Chol Song <nskystars at yahoo dot com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef ENABLE_SCROLLBACK

#include <assert.h>
#include <err.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cursor.h"
#include "font.h"
#include "framebuffer.h"
#include "scrollback.h"
#include "vtermlow.h"

#define DEFAULT_SCROLLBACK_HISTORY (256)

static struct {
	int history;
	int page;
	int queuePointer;
	bool queueFull;
} scrollback = {
	DEFAULT_SCROLLBACK_HISTORY,
	0,
	0,
	false
};

static void configHistory(const char *config);
static inline u_int historyCoordToIndex(TVterm *p, u_int x, u_int y);
static inline u_int coordToIndex(TVterm *p, u_int x, u_int y);
static void sb_move_to_hist(TVterm *p, int dst, int src, int n);
static int getQueueSize(void);
static int sb_queue_line_rel_to_abs(int rel);
static void historyUnclean(TVterm *p, u_int from, u_int to);
static void unclean(TVterm *p, u_int from, u_int to);
static void sb_draw_chars_hist(TVterm *p, u_int from, u_int to, u_int draw_offset);
static void sb_draw_chars(TVterm *p, u_int from, u_int to, u_int draw_offset);
static void refresh(TVterm *p);

static void configHistory(const char *config)
{
	int history;

	if (config != NULL) {
		history = atoi(config);
		if (history > 0 && history < 10000)
			scrollback.history = history;
		else
			warnx("Invalid scrollback history: %s", config);
	}
}

void scrollback_configure(TCaps *caps)
{
	const char *config;

	assert(caps != NULL);

	config = caps_findFirst(caps, "scrollback.history");
	configHistory(config);
}

int scrollback_getHistory(void)
{
	return scrollback.history;
}

/*
 * scrollback memory management
 */

static inline u_int historyCoordToIndex(TVterm *p, u_int x, u_int y)
{
	return (x + y * p->xcap4) % (p->xcap4 * scrollback.history);
}

static inline u_int coordToIndex(TVterm *p, u_int x, u_int y)
{
	return (x + y * p->xcap4) % p->tsize;
}

static void sb_move_to_hist(TVterm *p, int dst, int src, int n)
{
	memcpy(p->htext + dst, p->text + src, n * sizeof(u_short));
	memcpy(p->hlang + dst, p->lang + src, n);
#if 0 /* TODO: history copy and paste */
	memcpy(p->hrtxt + dst, p->rtxt + src, n * sizeof(u_short));
#endif
	memcpy(p->hfcol + dst, p->fcol + src, n);
	memcpy(p->hbcol + dst, p->bcol + src, n);
	memcpy(p->hflag + dst, p->flag + src, n);
}

/*
 * in-only circular queue
 */

static int getQueueSize(void)
{
	if (scrollback.queueFull)
		return scrollback.history;
	else
		return scrollback.queuePointer;
}

/* relative to absolute */
static int sb_queue_line_rel_to_abs(int relative)
{
	int result, queueSize;

	result = -1;
	queueSize = getQueueSize();
	if (relative >= 0 && relative <= queueSize)
		result = (scrollback.queuePointer - relative + scrollback.history) % scrollback.history;
	return result;
}

void scrollback_insertLine(TVterm *p, const int src)
{
	int dst;

	dst = historyCoordToIndex(p, 0, scrollback.queuePointer);
	sb_move_to_hist(p, dst, src, p->xcap4);
	scrollback.queuePointer++;

	scrollback.queuePointer %= scrollback.history;
	if (scrollback.queuePointer == scrollback.history - 1)
		scrollback.queueFull = true;
}

/*
 * scrollback refresh routine
 */

static void historyUnclean(TVterm *p, u_int from, u_int to)
{
	u_int x, y, i;

	for (y = from; y < to; y++) {
		for (x = 0; x < p->xcap; x++) {
			i = historyCoordToIndex(p, x, y);
			p->hflag[i] &= ~VTERM_FLAG_CLEAN;
		}
	}
}

static void unclean(TVterm *p, u_int from, u_int to)
{
	u_int x, y, i;

	for (y = from; y < to; y++) {
		for (x = 0; x < p->xcap; x++) {
			i = coordToIndex(p, x, y);
			p->flag[i] &= ~VTERM_FLAG_CLEAN;
		}
	}
}

static void sb_draw_chars_hist(TVterm *p, u_int from, u_int to, u_int draw_offset)
{
	TFont *font;
	u_int x, y, i, w;
	u_char hlang;

	for (y = from; y < to; y++) {
		for (x = 0; x < p->xcap; x++) {
			i = historyCoordToIndex(p, x, y);
			if (p->hflag[i] & VTERM_FLAG_CLEAN)
				continue; /* already clean */
			p->hflag[i] |= VTERM_FLAG_CLEAN;
			if (p->hflag[i] & VTERM_FLAG_CHAR_LEAD) {
				hlang = p->hlang[i];
				font = &(gFont[hlang]);
				p->hflag[i + 1] |= VTERM_FLAG_CLEAN;
				w = 2;
			} else if (p->hflag[i] & VTERM_FLAG_CHAR_TAIL)
				continue;
			else {
				hlang = p->hlang[i];
				font = &(gFont[hlang]);
				w = 1;
			}
			font_draw(font, p->htext[i], p->hfcol[i], p->hbcol[i],
				  x, ((y - from) + draw_offset), w);
		}
	}
}

static void sb_draw_chars(TVterm *p, u_int from, u_int to, u_int draw_offset)
{
	TFont *font;
	u_int x, y, i, w;
	u_char lang;

	for (y = from; y < to; y++) {
		for (x = 0; x < p->xcap; x++) {
			i = coordToIndex(p, x, y);
			if (p->flag[i] & VTERM_FLAG_CLEAN)
				continue; /* already clean */
			p->flag[i] |= VTERM_FLAG_CLEAN;
			if (p->flag[i] & VTERM_FLAG_CHAR_LEAD) {
				lang = p->lang[i];
				font = &(gFont[lang]);
				p->flag[i + 1] |= VTERM_FLAG_CLEAN;
				w = 2;
			} else if (p->flag[i] & VTERM_FLAG_CHAR_TAIL)
				continue;
			else {
				lang = p->lang[i];
				font = &(gFont[lang]);
				w = 1;
			}
			font_draw(font, p->text[i], p->fcol[i], p->bcol[i],
				  x, ((y - from) + draw_offset), w);
		}
	}
}

static void refresh(TVterm *p)
{
	int queueSize;
	int top, bottom;
	int top_abs, bottom_abs;
        sigset_t set, oldset;
 
	sigemptyset(&set);
	sigaddset(&set, SIGUSR1);
	sigprocmask(SIG_SETMASK, &set, &oldset);
 
	if (!p->active) {
		sigprocmask(SIG_SETMASK, &oldset, NULL);
		return;
	}

	cursor_show(p, &cursor, false);
	cursor_show(p, &mouseCursor, false);

	top = (p->ycap / 2) * scrollback.page;
	queueSize = getQueueSize();
	if (top > queueSize)
		top = queueSize;
	if (top < p->ycap)
		bottom = 0;
	else
		bottom = top - p->ycap;

	top_abs = sb_queue_line_rel_to_abs(top);
	bottom_abs = sb_queue_line_rel_to_abs(bottom);

#if 0
	fprintf(stderr, "page=%d,[top=%d,bottom=%d,top_abs=%d,bottom_abs=%d]\n",
		scrollback.page, top, bottom, top_abs, bottom_abs);
#endif

	if (top_abs >= bottom_abs) {
#if 0
		historyUnclean(p, top_abs, queueSize + 1);
#endif
		historyUnclean(p, top_abs, queueSize);
#if 0
		historyUnclean(p, 0, bottom_abs + 1);
#endif
		historyUnclean(p, 0, bottom_abs);

		sb_draw_chars_hist(p, top_abs, queueSize, 0);
#if 0
		sb_draw_chars_hist(p, 0, bottom_abs + 1, queueSize - top_abs);
#endif
		sb_draw_chars_hist(p, 0, bottom_abs, queueSize - top_abs);
	} else if (top_abs < bottom_abs) {
#if 0
		historyUnclean(p, top_abs, bottom_abs + 1);
#endif
		historyUnclean(p, top_abs, bottom_abs);
#if 0
		sb_draw_chars_hist(p, top_abs, bottom_abs + 1, 0);
#endif
		sb_draw_chars_hist(p, top_abs, bottom_abs, 0);
	}

	unclean(p, 0, (p->ycap > top) ? (p->ycap - top) : 0);
	sb_draw_chars(p, 0, (p->ycap > top) ? (p->ycap - top) : 0, top);
	unclean(p, 0, p->ycap);

	sigprocmask(SIG_SETMASK, &oldset, NULL);
}

/*
 * scrollback actions
 */

void sb_init_pageno(void)
{
	scrollback.page = 0;
}

void scrollback_backward(TVterm *p)
{
	int maxPages;

	maxPages = (getQueueSize() + (p->ycap / 2)) / (p->ycap / 2);
	if (maxPages < 0)
		maxPages = 0;
	if (maxPages > scrollback.page) {
		scrollback.page++;
		refresh(p);
	}
}

void scrollback_forward(TVterm *p)
{
	if (scrollback.page == 1) {
		scrollback.page = 0;
		unclean(p, 0, p->ycap);
#if 0
		sb_draw_chars(p, 0, p->ycap, 0);
#endif
		vterm_refresh(p);
	} else if (scrollback.page > 1) {
		scrollback.page--;
		refresh(p);
	}
}

bool sb_is_on(void)
{
	return scrollback.page != 0;
}

#endif /* ENABLE_SCROLLBACK */

