/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM Unofficial eXtension Patch
 * Copyright (C) 2006 Young Chol Song <nskystars at yahoo dot com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef ENABLE_SPLASH_SCREEN

#include <assert.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "framebuffer.h"
#include "palette.h"
#include "read-bmp.h"
#include "read-png.h"
#include "splash.h"
#include "utilities.h"

#define DEFAULT_SPLASH_TINT (5)

TSplash gSplash;

static TSplash *const self = &gSplash;
static bool initialized;

static struct {
	char *image;
	int tint;
} splash = {
	NULL,
	DEFAULT_SPLASH_TINT
};

static void finalize(void);
static void finalizer(void);
static void configImage(const char *config);
static void configTint(const char *config);
static void loadImage(void);
static void unloadImage(void);
static void applyTint(void);
static void resizeImage(u_int width, u_int height);

void splash_initialize(void)
{
	assert(!initialized);

	atexit(finalizer);
	initialized = true;
	self->bitmap = NULL;
	self->width  = 0;
	self->height = 0;
	bzero(self->mask, sizeof(self->mask));
	bzero(self->shift, sizeof(self->shift));
	self->enable = false;
}

static void finalize(void)
{
	assert(initialized);

	unloadImage();
	if (splash.image != NULL) {
		free(splash.image);
		splash.image = NULL;
	}
	initialized = false;
}

static void finalizer(void)
{
	finalize();
}

static void configImage(const char *config)
{
	assert(initialized);

	if (config != NULL) {
		if (splash.image != NULL)
			free(splash.image);
		splash.image = strdup(config);
		if (splash.image == NULL)
			err(1, "strdup()");
	}
}

static void configTint(const char *config)
{
	int tint;

	assert(initialized);

	if (config != NULL) {
		tint = atoi(config);
		if (tint >= 0 && tint <= 10)
			splash.tint = tint;
		else
			warnx("Invalid splash tint: %s", config);
	}
}

void splash_configure(TCaps *caps)
{
	const char *config;

	assert(initialized);
	assert(caps != NULL);

	config = caps_findFirst(caps, "splash.image");
	configImage(config);
	config = caps_findFirst(caps, "splash.tint");
	configTint(config);
}

static void loadImage(void)
{
	FILE *stream;

	assert(initialized);

	unloadImage();
	if (splash.image == NULL)
		return;
	if (strlen(splash.image) > 4 &&
	    strrcmp(splash.image, ".bmp") == 0) {
		if ((stream = fopen(splash.image, "r")) != NULL) {
			self->bitmap = read_bmp_file(stream,
						     &(self->width),
						     &(self->height));
			fclose(stream);
		} else
			warnx("Could not read bitmap file.");
	}
#ifdef WITH_LIBPNG
	else if (strlen(splash.image) > 4 &&
		 strrcmp(splash.image, ".png") == 0) {
		if ((stream = fopen(splash.image, "r")) != NULL) {
			self->bitmap = read_png_file(stream,
						     &(self->width),
						     &(self->height));
			fclose(stream);
		} else
			warnx("Could not read PNG file.");
	}
#endif
	if (self->bitmap != NULL)
		self->enable = true;
}

static void unloadImage(void)
{
	assert(initialized);

	self->enable = false;
	if (self->bitmap != NULL) {
		free(self->bitmap);
		self->bitmap = NULL;
	}
}

static void applyTint(void)
{
	int x, y;
	float t;

	assert(initialized);

	if (!self->enable)
		return;
	if (self->bitmap == NULL)
		return;
	t = ((10 - splash.tint) / 10.0f);
	for (y = 0; y < self->height; y++) {
		for (x = 0; x < self->width; x++) {
			SPLASH_GET(self->bitmap, self->width, x, y, 0) *= t;
			SPLASH_GET(self->bitmap, self->width, x, y, 1) *= t;
			SPLASH_GET(self->bitmap, self->width, x, y, 2) *= t;
		}
	}
}

static void resizeImage(u_int width, u_int height)
{
	int w, h;
	int x, y;
	u_char *bitmap;

	assert(initialized);

	if (!self->enable)
		return;
	if (self->bitmap == NULL)
		return;
	bitmap = (u_char *)malloc(height * width * 3 * sizeof(u_char));
	if (bitmap != NULL) {
		for (y = 0; y < height; y++) {
			h = y * self->height / height;
			for (x = 0; x < width; x++) {
				w = x * self->width / width;
				SPLASH_GET(bitmap, width, x, y, 0) =
					SPLASH_GET(self->bitmap, self->width,
						   w, h, 0);
				SPLASH_GET(bitmap, width, x, y, 1) =
					SPLASH_GET(self->bitmap, self->width,
						   w, h, 1);
				SPLASH_GET(bitmap, width, x, y, 2) =
					SPLASH_GET(self->bitmap, self->width,
						   w, h, 2);
			}
		}
	}
	free(self->bitmap);
	self->bitmap = bitmap;
	self->width = width;
	self->height = height;
}

void splash_load(void)
{
	int bpp, rlength, glength, blength;

	assert(initialized);

	bpp = gFramebuffer.cap.bitsPerPixel;
	if (bpp != 15 && bpp != 16 && bpp != 24 && bpp != 32 &&
	    splash.image != NULL) {
		warnx("Splash is not supported on %dbpp.", bpp);	
		return;
	}
	rlength = palette_getRedLength();
	glength = palette_getGreenLength();
	blength = palette_getBlueLength();
	self->mask[0] = (0xff << (8 - rlength));
	self->mask[1] = (0xff << (8 - glength));
	self->mask[2] = (0xff << (8 - blength));
	self->shift[2] = 8 - blength;
	self->shift[1] = glength - self->shift[2];
	self->shift[0] = rlength + self->shift[1];
	loadImage();
	applyTint();
	resizeImage(gFramebuffer.width, gFramebuffer.height);
}

#endif /* ENABLE_SPLASH_SCREEN */

