/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM/ME - J framebuffer terminal/Multilingual Enhancement -
 * Copyright (C) 2003 Fumitoshi UKAI <ukai@debian.or.jp>
 * Copyright (C) 1999 Noritoshi MASUICHI <nmasu@ma3.justnet.ne.jp>
 *
 * KON2 - Kanji ON Console -
 * Copyright (C) 1992,1993 Atusi MAEDA <mad@math.keio.ac.jp>
 * Copyright (C) 1992-1996 Takashi MANABE <manabe@papilio.tutics.tut.ac.jp>
 *
 * KON2 for FreeBSD
 * Copyright (C) Takashi OGURA
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <utmp.h>

#if defined (__linux__)
#include <sys/vt.h>
#include <pty.h>
#elif defined (__FreeBSD__)
#include <osreldate.h>
#if __FreeBSD_version >= 410000
#include <sys/consio.h>
#else
#include <machine/console.h>
#endif
#include <sys/ttycom.h>
#include <libutil.h>
#elif defined (__NetBSD__) || defined (__OpenBSD__)
#include <sys/ttycom.h>
#include <util.h>
#ifndef _DEV_WSCONS_WSDISPLAY_USL_IO_H_
#include <dev/wscons/wsdisplay_usl_io.h>
#define _DEV_WSCONS_WSDISPLAY_USL_IO_H_
#endif
#endif

#include "font.h"
#include "framebuffer.h"
#include "getcap.h"
#include "keyboard.h"
#include "main.h"
#include "mouse.h"
#include "privilege.h"
#include "screensaver.h"
#ifdef ENABLE_SCROLLBACK
#include "scrollback.h"
#endif
#include "term.h"
#include "utilities.h"
#include "vterm.h"
#include "vtermlow.h"

static TTerm term;
static TTerm *const self = &term;
static pid_t child;
static fd_set orgReadFds;
static int numFds;
static bool initialized;

typedef enum {
	TERMIOS,
	VT_MODE
} SAVE_STATUS;
static bool saved[2];

static void finalize(void);
static void finalizer(void);
static void childHandler(int signum);
static bool openPseudoTerminal(void);
static void executeShell(const char *shell, char *const args[],
			 const char *name);
static void consoleHandler(void);
static void writeLoginRecord(void);
static void eraseLoginRecord(void);
#if defined (__FreeBSD__)
static int sysconsTolinux(u_char *buf, const int len, const int max);
#endif
#if defined (__NetBSD__) || defined (__OpenBSD__)
static int wsconsTolinux(u_char *buf, const int len, const int max);
#endif

void term_initialize(TCaps *caps, const char *encoding)
{
	struct sigaction act;

	assert(!initialized);
	assert(caps != NULL);
	assert(encoding != NULL);

	atexit(finalizer);
	initialized = true;
	self->masterPty = -1;
	self->slavePty  = -1;
	self->device[0] = '\0';
	if (tcgetattr(STDIN_FILENO, &(self->termios)) == -1)
		err(1, "tcgetattr()");
	saved[TERMIOS] = true;
	if (ioctl(STDIN_FILENO, VT_GETMODE, &(self->vt_mode)) == -1)
		err(1, "ioctl(VT_GETMODE)");
	saved[VT_MODE] = true;
	if (!openPseudoTerminal())
		errx(1, "Could not open pseudo-terminal.");
	vterm_initialize(&(self->vterm), self, caps,
			 gFramebuffer.width / gFontsWidth,
			 gFramebuffer.height / gFontsHeight,
			 encoding);
	bzero(&act, sizeof(act));
	act.sa_handler = childHandler;
	act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
	sigaction(SIGCHLD, &act, NULL);
}

static void finalize(void)
{
	struct sigaction act;

	assert(initialized);

	eraseLoginRecord();
	vterm_finalize(&(self->vterm));
	if (saved[TERMIOS])
		if (tcsetattr(STDIN_FILENO, TCSANOW, &(self->termios)) == -1)
			warn("tcsetattr()");
	if (saved[VT_MODE])
		if (ioctl(STDIN_FILENO, VT_SETMODE, &(self->vt_mode)) == -1)
			warn("ioctl(VT_SETMODE)");
	if (self->slavePty != -1) {
		close(self->slavePty);
		self->slavePty = -1;
	}
	if (self->masterPty != -1) {
		close(self->masterPty);
		self->masterPty = -1;
	}
	bzero(&act, sizeof(act));
	act.sa_handler = SIG_DFL;
	sigaction(SIGCHLD, &act, NULL);
	initialized = false;
}

static void finalizer(void)
{
	finalize();
}

static void childHandler(int signum)
{
	int status, errsv;
	pid_t pid;

	errsv = errno;
	pid = waitpid(child, &status, WNOHANG);
	if (pid == (pid_t)-1 && errno != ECHILD)
		err(1, "waitpid()");
	if (pid == child || (pid == (pid_t)-1 && errno == ECHILD)) {
		if (WIFSIGNALED(status))
			errx(1, "Child process died with signal %s",
			     sys_siglist[WTERMSIG(status)]);
		if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
			errx(1, "Child process exited with status %d",
			     WEXITSTATUS(status));
		fputs("Finished.\n", stderr);
		exit(EXIT_SUCCESS);
	}
	errno = errsv;
}

static bool openPseudoTerminal(void)
{
	int status;
	char *device;

	assert(initialized);

	/* openpty(3) don't use thrid argument. */
	privilege_on();
	status = openpty(&self->masterPty, &self->slavePty, NULL, NULL, NULL);
	privilege_off();
	if (status == -1)
		return false;
	device = ttyname(self->slavePty);
	if (device == NULL)
		return false;
	strncpy(self->device, device, MAXPATHLEN);
	self->device[MAXPATHLEN - 1] = '\0';
	return true;
}

void term_mouseEnable(void)
{
	int mouseFd;

	assert(initialized);

	mouseFd = mouse_getFd();
	if (mouseFd != -1)
		FD_SET(mouseFd, &orgReadFds);
	if (mouseFd > self->masterPty)
		numFds = mouseFd + 1;
	else
		numFds = self->masterPty + 1;
}

void term_mouseDisable(void)
{
	int mouseFd;

	assert(initialized);

	mouseFd = mouse_getFd();
	if (mouseFd != -1)
		FD_CLR(mouseFd, &orgReadFds);
	numFds = self->masterPty + 1;
}

void term_execute(const char *shell, char *const args[], const char *name)
{
	struct termios termios;
	int on;

	assert(initialized);
	assert(shell != NULL);
	assert(args != NULL);
	assert(name != NULL);

	fflush(stdout);
	fflush(stderr);
	child = fork();
	if (child == (pid_t)-1)
		err(1, "fork()");
	else if (child == (pid_t)0) {
		/* Child process */
		executeShell(shell, args, name);
	} else {
		/* Parent process */
		termios = self->termios;
#if defined (__linux__)
		termios.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN | XCASE);
#elif defined (__FreeBSD__) || defined (__NetBSD__) || defined (__OpenBSD__)
		termios.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
#elif
	#error not implement
#endif
		termios.c_iflag = 0;
		termios.c_oflag = (OPOST | ONLCR);
		termios.c_cc[VMIN] = 1;
		termios.c_cc[VTIME] = 0;
#if defined (__FreeBSD__) || defined (__NetBSD__) || defined (__OpenBSD__)
		termios.c_cc[VDISCARD] = _POSIX_VDISABLE;
		termios.c_cc[VLNEXT] = _POSIX_VDISABLE;
		termios.c_cc[VSTART] = _POSIX_VDISABLE;
		termios.c_cc[VSTOP] = _POSIX_VDISABLE;
		termios.c_cc[VINTR] = _POSIX_VDISABLE;
		termios.c_cc[VSUSP] = _POSIX_VDISABLE;
		termios.c_cc[VDSUSP] = _POSIX_VDISABLE;
		termios.c_cc[VQUIT] = _POSIX_VDISABLE;
#endif
		termios.c_cflag &= ~(CSIZE | PARENB);
		termios.c_cflag |= CS8;
#if defined (__linux__)
		termios.c_line = 0;
#endif
		if (tcsetattr(STDIN_FILENO, TCSANOW, &termios) == -1)
			err(1, "tcsetattr()");
		on = 1;
		if (ioctl(self->masterPty, TIOCPKT, &on) == -1)
			err(1, "ioctl(TIOCPKT)");
		consoleHandler();
	}
	/* NOTREACHED */
}

static void executeShell(const char *shell, char *const args[],
			 const char *name)
{
	int i;

	assert(initialized);
	assert(shell != NULL);
	assert(args != NULL);
	assert(name != NULL);

	setenv("TERM", name, 1);
	close(self->masterPty);
	self->masterPty = -1;
	if (login_tty(self->slavePty) == -1)
		err(1, "login_tty()");
	if (tcsetattr(STDIN_FILENO, TCSANOW, &(self->termios)) == -1)
		err(1, "tcsetattr()");
	writeLoginRecord();
	privilege_drop();
	for (i = getdtablesize(); i > STDERR_FILENO; i--)
		close(i);
	execvp(shell, args);
	exit(EXIT_FAILURE);
	/* NOTREACHED */
}

#define BUF_SIZE (1024)

static void consoleHandler(void)
{
	int i, v;
	struct timeval tv;
	char buf[BUF_SIZE + 1];
	fd_set readFds;

	assert(initialized);

	FD_ZERO(&orgReadFds);
	FD_SET(STDIN_FILENO, &orgReadFds);
	FD_SET(self->masterPty, &orgReadFds);
	if (mouse_isEnable())
		mouse_start();
	if (mouse_getFd() != -1)
		term_mouseEnable();
	else
		term_mouseDisable();
	for (;;) {
		do {
			/* Idle loop. */
			vterm_pollCursor(&(self->vterm), false);
			readFds = orgReadFds;
			tv.tv_sec  = 0;
			tv.tv_usec = 100000; /* 0.1 sec. */
			v = select(numFds, &readFds, NULL, NULL, &tv);
		} while (v == 0 || (v == -1 && (errno == EINTR || mouse_getFd() == -1)));
		if (v == -1)
			err(1, "select()");
		if (FD_ISSET(STDIN_FILENO, &readFds)) {
			i = read(STDIN_FILENO, buf, BUF_SIZE);
#if defined (__FreeBSD__)
			i = sysconsTolinux(buf, i, BUF_SIZE);
#elif defined (__NetBSD__) || defined (__OpenBSD__)
			i = wsconsTolinux(buf, i, BUF_SIZE);
#endif
			if (i > 0) {
#ifdef ENABLE_SCROLLBACK
				/* Scroll Lock + Page Up */
				bool scrollLocked;
				scrollLocked = keyboard_isScrollLocked();
				if (scrollLocked &&
				    buf[0] == '\033' && buf[1] == '[' &&
				    buf[2] == '5' && buf[3] == '~') {
					scrollback_backward(&(self->vterm));
					continue;
				}
				/* Scroll Lock + Page Down */
				else if (scrollLocked &&
					 buf[0] == '\033' && buf[1] == '[' &&
					 buf[2] == '6' && buf[3] == '~') {
					scrollback_forward(&(self->vterm));
					continue;
				}
#endif
				write_wrapper(self->masterPty, buf, i);
			}
			vterm_pollCursor(&(self->vterm), true);
		}
		if (FD_ISSET(self->masterPty, &readFds)) {
#ifdef ENABLE_SCROLLBACK
			bool scrollLocked;
			scrollLocked = keyboard_isScrollLocked();
			if (scrollLocked)
				usleep(10000); /* 0.01 sec. */
			else {
				i = read(self->masterPty, buf, BUF_SIZE);
				if (i > 0) {
					sb_init_pageno();
					vterm_emulate(&(self->vterm), buf, i);
					vterm_refresh(&(self->vterm));
				}
			}
#else
			i = read(self->masterPty, buf, BUF_SIZE);
			if (i > 0) {
				vterm_emulate(&(self->vterm), buf, i);
				vterm_refresh(&(self->vterm));
			}
#endif
		}
		if (mouse_isEnable() && mouse_getFd() != -1) {
			if (FD_ISSET(mouse_getFd(), &readFds) &&
			    ((&self->vterm))->active) {
				mouse_getPackets(&(self->vterm));
				vterm_pollCursor(&(self->vterm), true);
			}
		}
	}
	/* NOTREACHED */
}

static void writeLoginRecord(void)
{
#if defined (__linux__)
	struct utmp utmp;
	struct passwd *pw;
	char *tn;

	assert(initialized);

	bzero(&utmp, sizeof(utmp));
	pw = getpwuid(privilege_getUID());
	if (pw == NULL)
		return; /* give up */
	if (strncmp(self->device, "/dev/pts/", 9) == 0) {
		/* Unix98 style: "/dev/pts/\*" */
		tn = self->device + 9;
		if (strlen(tn) < 1)
			return; /* bad format */
		snprintf(utmp.ut_id, sizeof(utmp.ut_id), "p%s", tn);
	} else if (strncmp(self->device, "/dev/tty", 8) == 0) {
		/* BSD style: "/dev/tty[p-za-e][0-9a-f]" */
		tn = self->device + 8;
		if (strlen(tn) < 2)
			return; /* bad format */
		strncpy(utmp.ut_id, tn, sizeof(utmp.ut_id));
	} else
		return; /* bad format */
	utmp.ut_type = DEAD_PROCESS;
	privilege_on();
	setutent();
	getutid(&utmp);
	utmp.ut_type = USER_PROCESS;
	utmp.ut_pid = getpid();
	tn = self->device + 5;
	strncpy(utmp.ut_line, tn, sizeof(utmp.ut_line));
	strncpy(utmp.ut_user, pw->pw_name, sizeof(utmp.ut_user));
	time(&(utmp.ut_time));
	pututline(&utmp);
	endutent();
	privilege_off();
#elif defined (__FreeBSD__) || defined (__NetBSD__) || defined (__OpenBSD__)
	struct utmp utmp;
	struct passwd *pw;
	char *tn;
	int fd, ts;

	assert(initialized);

	bzero(&utmp, sizeof(utmp));
	pw = getpwuid(privilege_getUID());
	if (pw == NULL)
		return; /* give up */
	if (strncmp(self->device, "/dev/tty", 8) == 0) {
		/* BSD style: "/dev/tty[p-sP-S][0-9a-v]" */
		tn = self->device + 5;
		if (strlen(tn) < 5)
			return; /* bad format */
	} else
		return; /* bad format */
	strncpy(utmp.ut_line, tn, sizeof(utmp.ut_line));
	strncpy(utmp.ut_name, pw->pw_name, sizeof(utmp.ut_name));
	time(&(utmp.ut_time));
	if ((ts = ttyslot()) == 0)
		return; /* give up */
	privilege_on();
	fd = open(_PATH_UTMP, O_WRONLY | O_CREAT, 0644);
	privilege_off();
	if (fd != -1) {
		lseek(fd, (ts * sizeof(struct utmp)), L_SET);
		write(fd, &utmp, sizeof(struct utmp));
		close(fd);
	}
#else
	#error not implement
#endif
}

static void eraseLoginRecord(void)
{
#if defined (__linux__)
	struct utmp utmp, *utp;
	char *tn;

	assert(initialized);

	bzero(&utmp, sizeof(utmp));
	if (strncmp(self->device, "/dev/pts/", 9) == 0) {
		/* Unix98 style: "/dev/pts/\*" */
		tn = self->device + 9;
		if (strlen(tn) < 1)
			return; /* bad format */
		snprintf(utmp.ut_id, sizeof(utmp.ut_id), "p%s", tn);
	} else if (strncmp(self->device, "/dev/tty", 8) == 0) {
		/* BSD style: "/dev/tty[p-za-e][0-9a-f]" */
		tn = self->device + 8;
		if (strlen(tn) < 2)
			return; /* bad format */
		strncpy(utmp.ut_id, tn, sizeof(utmp.ut_id));
	} else
		return; /* bad format */
	utmp.ut_type = USER_PROCESS;
	privilege_on();
	setutent();
	utp = getutid(&utmp);
	if (utp == NULL) {
		endutent();
		privilege_off();
		return; /* give up */
	}
	utp->ut_type = DEAD_PROCESS;
	bzero(utp->ut_user, sizeof(utmp.ut_user));
	time(&(utp->ut_time));
	pututline(utp);
	endutent();
	privilege_off();
#elif defined (__FreeBSD__) || defined (__NetBSD__) || defined (__OpenBSD__)
	struct utmp utmp;
	char *tn;
	int fd, ts;

	assert(initialized);

	if (strncmp(self->device, "/dev/tty", 8) == 0) {
		/* BSD style: "/dev/tty[p-sP-S][0-9a-v]" */
		tn = self->device + 5;
		if (strlen(tn) < 5)
			return; /* bad format */
	} else
		return; /* bad format */
	privilege_on();
	fd = open(_PATH_UTMP, O_RDWR);
	privilege_off();
	if (fd != -1) {
		for (ts = 0; read(fd, &utmp, sizeof(struct utmp)) ==
		     sizeof(struct utmp); ts++) {
			if (strncmp(utmp.ut_line, tn, strlen(tn)) == 0) {
				bzero(utmp.ut_name, sizeof(utmp.ut_name));
				bzero(utmp.ut_host, sizeof(utmp.ut_host));
				time(&utmp.ut_time);
				lseek(fd, (ts * sizeof(struct utmp)), L_SET);
				write(fd, &utmp, sizeof(struct utmp));
				break;
			}
		}
		close(fd);
	}
#else
	#error not implement
#endif
}

#if defined (__FreeBSD__)
static int sysconsTolinux(u_char *buf, const int len, const int max)
{
	u_char *p;
	int i, newlen;

	assert(initialized);
	assert(buf != NULL);

	newlen = len;
	for (i = 0, p = buf; i < newlen; i++, p++) {
		switch (*p) {
		case '\177':      /* Delete */
			newlen += strrep(p, len - i, max - i, 1, "\033[3~");
			i += 3; p += 3;
			break;
		case '\033':      /* ESC */
			if (i + 2 <= newlen && *(p + 1) == '[') {
				switch (*(p + 2)) {
				case 'E':        /* Center (Num Lock + 5) */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[G");
					i += 2; p += 2;
					break;
				case 'H':        /* Home */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[1~");
					i += 3; p += 3;
					break;
				case 'L':        /* Insert */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[2~");
					i += 3; p += 3;
					break;
				case 'F':        /* End */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[4~");
					i += 3; p += 3;
					break;
				case 'I':        /* Page Up */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[5~");
					i += 3; p += 3;
					break;
				case 'G':        /* Page Down */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[6~");
					i += 3; p += 3;
					break;
				case 'M':        /* F1 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[[A");
					i += 3; p += 3;
					break;
				case 'N':        /* F2 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[[B");
					i += 3; p += 3;
					break;
				case 'O':        /* F3 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[[C");
					i += 3; p += 3;
					break;
				case 'P':        /* F4 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[[D");
					i += 3; p += 3;
					break;
				case 'Q':        /* F5 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[[E");
					i += 3; p += 3;
					break;
				case 'R':        /* F6 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[17~");
					i += 4; p += 4;
					break;
				case 'S':        /* F7 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[18~");
					i += 4; p += 4;
					break;
				case 'T':        /* F8 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[19~");
					i += 4; p += 4;
					break;
				case 'U':        /* F9 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[20~");
					i += 4; p += 4;
					break;
				case 'V':        /* F10 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[21~");
					i += 4; p += 4;
					break;
				case 'W':        /* F11 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[23~");
					i += 4; p += 4;
					break;
				case 'X':        /* F12 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[24~");
					i += 4; p += 4;
					break;
				case 'Y':        /* F13 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[25~");
					i += 4; p += 4;
					break;
				case 'Z':        /* F14 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[26~");
					i += 4; p += 4;
					break;
				case 'a':        /* F15 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[28~");
					i += 4; p += 4;
					break;
				case 'b':        /* F16 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[29~");
					i += 4; p += 4;
					break;
				case 'c':        /* F17 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[31~");
					i += 4; p += 4;
					break;
				case 'd':        /* F18 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[32~");
					i += 4; p += 4;
					break;
				case 'e':        /* F19 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[33~");
					i += 4; p += 4;
					break;
				case 'f':        /* F20 */
					newlen += strrep(p, len - i, max - i,
							 3, "\033[34~");
					i += 4; p += 4;
					break;
				default:
					i += 2; p += 2;
					break;
				}
			}
		default:
			break;
		}
	}
	return newlen;
}
#endif

#if defined (__NetBSD__) || defined (__OpenBSD__)
static int wsconsTolinux(u_char *buf, const int len, const int max)
{
	u_char *p;
	int i, n, newlen;

	assert(initialized);
	assert(buf != NULL);

	newlen = len;
	for (i = 0, p = buf; i < newlen; i++, p++) {
		switch (*p) {
#if defined (__NetBSD__)
		case '\177':      /* Delete */
			newlen += strrep(p, len - i, max - i, 1, "\033[3~");
			i += 3; p += 3;
			break;
#endif
		case '\033':      /* ESC */
			if (i + 3 <= newlen && *(p + 1) == '[' &&
			    *(p + 3) == '~') {
				switch (*(p + 2)) {
				case '7':        /* Home */
					newlen += strrep(p, len - i, max - i,
							 4, "\033[1~");
					i += 3; p += 3;
					break;
				case '8':        /* End */
					newlen += strrep(p, len - i, max - i,
							 4, "\033[4~");
					i += 3; p += 3;
					break;
				default:
					i += 3; p += 3;
					break;
				}
			} else if (i + 4 <= newlen && *(p + 1) == '[' &&
			    isdigit(*(p + 2)) && isdigit(*(p + 3)) &&
			    *(p + 4) == '~') {
				n = hex2int(*(p + 2)) * 10 + hex2int(*(p + 3));
				switch (n) {
				case 11:         /* F1 */
					newlen += strrep(p, len - i, max - i,
							 5, "\033[[A");
					i += 3; p += 3;
					break;
				case 12:         /* F2 */
					newlen += strrep(p, len - i, max - i,
							 5, "\033[[B");
					i += 3; p += 3;
					break;
				case 13:         /* F3 */
					newlen += strrep(p, len - i, max - i,
							 5, "\033[[C");
					i += 3; p += 3;
					break;
				case 14:         /* F4 */
					newlen += strrep(p, len - i, max - i,
							 5, "\033[[D");
					i += 3; p += 3;
					break;
				case 15:         /* F5 */
					newlen += strrep(p, len - i, max - i,
							 5, "\033[[E");
					i += 3; p += 3;
					break;
				default:
					i += 4; p += 4;
					break;
				}
			}
		default:
			break;
		}
	}
	return newlen;
}
#endif

