/*
 * JFBTERM for FreeBSD
 * Copyright (C) 2008 Yusuke BABA <babayaga1@y8.dion.ne.jp>
 *
 * JFBTERM/ME - J framebuffer terminal/Multilingual Enhancement -
 * Copyright (C) 2003 Fumitoshi UKAI <ukai@debian.or.jp>
 * Copyright (C) 1999 Noritoshi MASUICHI <nmasu@ma3.justnet.ne.jp>
 *
 * KON2 - Kanji ON Console -
 * Copyright (C) 1992,1993 Atusi MAEDA <mad@math.keio.ac.jp>
 * Copyright (C) 1992-1996 Takashi MANABE <manabe@papilio.tutics.tut.ac.jp>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUSUKE BABA ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL YUSUKE BABA BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined (__linux__) || defined (__NetBSD__)
#define LIBICONV_PLUG
#endif

#if defined (__GLIBC__)
#define __USE_STRING_INLINES
#endif

#include <sys/types.h>
#include <sys/ioctl.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <iconv.h>
#include <signal.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>

#if defined (__linux__)
#include <sys/vt.h>
#elif defined (__FreeBSD__)
#include <osreldate.h>
#if __FreeBSD_version >= 410000
#include <sys/consio.h>
#else
#include <machine/console.h>
#endif
#elif defined (__NetBSD__) || defined (__OpenBSD__)
#ifndef _DEV_WSCONS_WSDISPLAY_USL_IO_H_
#include <dev/wscons/wsdisplay_usl_io.h>
#define _DEV_WSCONS_WSDISPLAY_USL_IO_H_
#endif
#endif

#include "clipboard.h"
#include "cursor.h"
#include "font.h"
#include "framebuffer.h"
#include "keyboard.h"
#include "mouse.h"
#include "palette.h"
#include "screensaver.h"
#ifdef ENABLE_SCROLLBACK
#include "scrollback.h"
#endif
#include "term.h"
#include "utilities.h"
#include "vterm.h"
#include "vtermlow.h"

static TVterm *object = NULL;

static inline u_int coordToIndex(TVterm *p, u_int x, u_int y);
static inline bool isLeadChar(TVterm *p, u_int x, u_int y);
static inline bool isTailChar(TVterm *p, u_int x, u_int y);
static inline void adjustCoord(TVterm *p, u_int *x, u_int *y);
static inline void brmove(void *dst, void *src, int n);
static inline void unclean(void *head, int n);
static inline void unclean4(void *head, int n);
static inline void vterm_move(TVterm *p, int dst, int src, int n);
static inline void vterm_brmove(TVterm *p, int dst, int src, int n);
static inline void vterm_clear(TVterm *p, int top, int n);
static void leaveVirtualTerminal(int signum);
static void enterVirtualTerminal(int signum);
static void vterm_text_clean_band(TVterm *p, u_int top, u_int bottom);
static void pollMouseCursor(TVterm *p);

static inline u_int coordToIndex(TVterm *p, u_int x, u_int y)
{
	return (x + y * p->xcap4) % p->tsize;
}

static inline bool isLeadChar(TVterm *p, u_int x, u_int y)
{
	return p->flag[coordToIndex(p, x, y)] & VTERM_FLAG_CHAR_LEAD;
}

static inline bool isTailChar(TVterm *p, u_int x, u_int y)
{
	return p->flag[coordToIndex(p, x, y)] & VTERM_FLAG_CHAR_TAIL;
}

static inline void adjustCoord(TVterm *p, u_int *x, u_int *y)
{
	if (isTailChar(p, *x, *y))
		-- * x;
}

static inline void brmove(void *dst, void *src, int n)
{
	memmove((void *)((char *)dst - n), (void *)((char *)src - n), n);
}

static inline void unclean(void *head, int n)
{
	unsigned char *c, *e;

	c = (unsigned char *)head;
	e = (unsigned char *)head + n;
	for (; c < e; c++)
		*c &= ~VTERM_FLAG_CLEAN;
}

static inline void unclean4(void *head, int n)
{
	unsigned int *l, *e;

	l = (unsigned int *)head;
	e = (unsigned int *)head + (n >> 2);
	for (; l < e; l++)
		*l &= 0x7f7f7f7f; /* ~VTERM_FLAG_CLEAN */
}

static inline void vterm_move(TVterm *p, int dst, int src, int n)
{
	memmove(p->text + dst, p->text + src, n * sizeof(u_short));
	memmove(p->lang + dst, p->lang + src, n);
	memmove(p->rtxt + dst, p->rtxt + src, n * sizeof(u_short));
	memmove(p->fcol + dst, p->fcol + src, n);
	memmove(p->bcol + dst, p->bcol + src, n);
	memmove(p->flag + dst, p->flag + src, n);
}

static inline void vterm_brmove(TVterm *p, int dst, int src, int n)
{
	brmove(p->text + dst, p->text + src, n * sizeof(u_short));
	brmove(p->lang + dst, p->lang + src, n);
	brmove(p->rtxt + dst, p->rtxt + src, n * sizeof(u_short));
	brmove(p->fcol + dst, p->fcol + src, n);
	brmove(p->bcol + dst, p->bcol + src, n);
	brmove(p->flag + dst, p->flag + src, n);
}

static inline void vterm_clear(TVterm *p, int top, int n)
{
	bzero(p->text + top, n * sizeof(u_short));
	bzero(p->lang + top, n);
	bzero(p->rtxt + top, n * sizeof(u_short));
	bzero(p->flag + top, n);
	memset(p->fcol + top, p->pen.fcol, n);
	memset(p->bcol + top, p->pen.bcol, n);
}

void vterm_delete_n_chars(TVterm *p, int n)
{
	u_int i;
	u_int dx;

	i = coordToIndex(p, p->pen.x, p->pen.y);
	dx = p->xcap - p->pen.x - n;

	vterm_move(p, i, i + n, dx); /* bmove */
	unclean(p->flag + i, dx);

	i = coordToIndex(p, p->xcap - n, p->pen.y);
	vterm_clear(p, i, n);
}

void vterm_insert_n_chars(TVterm *p, int n)
{
	u_int i;
	u_int dx;

	i = coordToIndex(p, p->xcap - 1, p->pen.y);
	dx = p->xcap - p->pen.x - n;
	vterm_brmove(p, i, i - n, dx);

	i = coordToIndex(p, p->pen.x, p->pen.y);
	unclean(p->flag + i + n, dx);
	vterm_clear(p, i, n);
}

void vterm_refresh(TVterm *p)
{
	TFont *font;
	u_int i, x, y, w;
	u_char lang;
	sigset_t set, oldset;

	sigemptyset(&set);
	sigaddset(&set, SIGUSR1);
	sigprocmask(SIG_SETMASK, &set, &oldset); 

	if (!p->active) {
		sigprocmask(SIG_SETMASK, &oldset, NULL); 
		return;
	}

	cursor_show(p, &cursor, false);
	cursor_show(p, &mouseCursor, false);

	if (p->textClear) {
		gFramebuffer.cap.fill(&gFramebuffer,
				      0,
				      0,
				      gFramebuffer.width,
				      gFramebuffer.height,
				      0);
		p->textClear = false;
	}
	for (y = 0; y < p->ycap; y++) {
		for (x = 0; x < p->xcap; x++) {
			i = coordToIndex(p, x, y);
			if (p->flag[i] & VTERM_FLAG_CLEAN)
				continue; /* already clean */
			p->flag[i] |= VTERM_FLAG_CLEAN;
			if (p->flag[i] & VTERM_FLAG_CHAR_LEAD) {
				lang = p->lang[i];
				font = &(gFont[lang]);
				p->flag[i + 1] |= VTERM_FLAG_CLEAN;
				w = 2;
			} else if (p->flag[i] & VTERM_FLAG_CHAR_TAIL)
				continue;
			else {
				lang = p->lang[i];
				font = &(gFont[lang]);
				w = 1;
			}
			font_draw(font, p->text[i], p->fcol[i], p->bcol[i],
				  x, y, w);
		}
	}
	/* XXX: pen position go out of screen by resize(1) for example */
	if (p->pen.x < p->xcap && p->pen.y < p->ycap) {
		cursor.wide = isLeadChar(p, p->pen.x, p->pen.y);
		cursor.x = p->pen.x;
		cursor.y = p->pen.y;
		cursor_show(p, &cursor, true);
	}

	sigprocmask(SIG_SETMASK, &oldset, NULL); 
}

void vterm_unregister_signal(void)
{
	struct sigaction act_sigusr1, act_sigusr2;

	if (object == NULL)
		return;

	bzero(&act_sigusr1, sizeof(act_sigusr1));
	act_sigusr1.sa_handler = SIG_DFL;
	sigaction(SIGUSR1, &act_sigusr1, NULL);

	bzero(&act_sigusr2, sizeof(act_sigusr2));
	act_sigusr2.sa_handler = SIG_DFL;
	sigaction(SIGUSR2, &act_sigusr2, NULL);

	ioctl(STDIN_FILENO, VT_RELDISP, 1); /* 1 = VT_TRUE */
}

void vterm_register_signal(TVterm *p)
{
	struct sigaction act_sigusr1, act_sigusr2;
	struct vt_mode vt_mode;

	object = p;

	ioctl(STDIN_FILENO, VT_RELDISP, VT_ACKACQ);

	bzero(&act_sigusr1, sizeof(act_sigusr1));
	act_sigusr1.sa_handler = leaveVirtualTerminal;
	act_sigusr1.sa_flags = SA_RESTART;
	sigaction(SIGUSR1, &act_sigusr1, NULL);

	bzero(&act_sigusr2, sizeof(act_sigusr2));
	act_sigusr2.sa_handler = enterVirtualTerminal;
	act_sigusr2.sa_flags = SA_RESTART;
	sigaction(SIGUSR2, &act_sigusr2, NULL);

	vt_mode.mode   = VT_PROCESS;
	vt_mode.waitv  = 0;
	vt_mode.relsig = SIGUSR1;
	vt_mode.acqsig = SIGUSR2;
	vt_mode.frsig  = SIGUSR1;
	if (ioctl(STDIN_FILENO, VT_SETMODE, &vt_mode) == -1)
		warn("ioctl(VT_SETMODE)");

	unclean4(p->flag, p->tsize);
	p->textClear = true;
	vterm_refresh(p);
}

static void leaveVirtualTerminal(int signum)
{
	int errsv;

	errsv = errno;
	object->active = false;
	if (mouse_isEnable()) {
		term_mouseDisable(object->term);
		mouse_stop();
	}
#if defined (__NetBSD__) || defined (__OpenBSD__)
	palette_restore(); /* maybe bugs in wscons(4). */
#endif
#ifdef ENABLE_SCROLLBACK
	sb_init_pageno();
#endif
	ioctl(STDIN_FILENO, VT_RELDISP, 1); /* 1 = VT_TRUE */
	errno = errsv;
}

static void enterVirtualTerminal(int signum)
{
	int errsv;

	errsv = errno;
	if (!object->active) {
		object->active = true;
		vterm_register_signal(object);
		if (mouse_isEnable()) {
			mouse_start();
			term_mouseEnable(object->term);
		}
		screensaver_execute(false);
		palette_reset();
	}
	errno = errsv;
}

void vterm_wput(TVterm *p, u_int idx, u_char ch1, u_char ch2, u_char raw1, u_char raw2)
{
	u_int i;

	i = coordToIndex(p, p->pen.x, p->pen.y);
	p->fcol[i] = p->pen.fcol;
	p->bcol[i] = p->pen.bcol;
	p->text[i] = (ch1 << 8) | ch2;
	p->lang[i] = idx;
	p->rtxt[i] = (raw1 << 8) | raw2;
	p->flag[i] = VTERM_FLAG_CHAR_LEAD | VTERM_FLAG_MULTIBYTE;
	i++;
	p->text[i] = 0; /* unused. */
	p->lang[i] = 0; /* unused. */
	p->rtxt[i] = 0; /* unused. */
	p->flag[i] = VTERM_FLAG_CHAR_TAIL | VTERM_FLAG_MULTIBYTE;
}

void vterm_sput(TVterm *p, u_int idx, u_char ch, u_char raw)
{
	u_int i;

	i = coordToIndex(p, p->pen.x, p->pen.y);
	p->fcol[i] = p->pen.fcol;
	p->bcol[i] = p->pen.bcol;
	p->text[i] = ch;
	p->lang[i] = idx;
	p->rtxt[i] = raw;
	p->flag[i] = VTERM_FLAG_CHAR_SINGLE | VTERM_FLAG_MULTIBYTE;
}

#ifdef ENABLE_UTF8
void vterm_uput1(TVterm *p, u_int idx, u_short ch, u_short raw)
{
	u_int i;

	i = coordToIndex(p, p->pen.x, p->pen.y);
	p->fcol[i] = p->pen.fcol;
	p->bcol[i] = p->pen.bcol;
	p->text[i] = ch;
	p->lang[i] = idx;
	p->rtxt[i] = raw;
	p->flag[i] = VTERM_FLAG_CHAR_SINGLE | VTERM_FLAG_UNICODE;
}

void vterm_uput2(TVterm *p, u_int idx, u_short ch, u_short raw)
{
	u_int i;

	i = coordToIndex(p, p->pen.x, p->pen.y);
	p->fcol[i] = p->pen.fcol;
	p->bcol[i] = p->pen.bcol;
	p->text[i] = ch;
	p->lang[i] = idx;
	p->rtxt[i] = raw;
	p->flag[i] = VTERM_FLAG_CHAR_LEAD | VTERM_FLAG_UNICODE;
	i++;
	p->text[i] = 0; /* unused. */
	p->lang[i] = 0; /* unused. */
	p->rtxt[i] = 0; /* unused. */
	p->flag[i] = VTERM_FLAG_CHAR_TAIL | VTERM_FLAG_UNICODE;
}
#endif

void vterm_text_clear_all(TVterm *p)
{
	u_int y;
	u_int i;

	for (y = 0; y < p->ymax; y++) {
		i = coordToIndex(p, 0, y);
		vterm_clear(p, i, p->xcap4); /* lclear */
	}
	p->textClear = true;
}

void vterm_text_clear_eol(TVterm *p, int mode)
{
	u_int i;
	u_int len;
	u_int x;

	switch (mode) {
	case 1:
		x = 0;
		len = p->pen.x;
		break;
	case 2:
		x = 0;
		len = p->xcap;
		break;
	default:
		x = p->pen.x;
		len = p->xcap - p->pen.x;
		break;
	}
	i = coordToIndex(p, x, p->pen.y);
	vterm_clear(p, i, len);
}

void vterm_text_clear_eos(TVterm *p, int mode)
{
	u_int i;
	u_int len;

	switch (mode) {
	case 1:
		vterm_text_clean_band(p, 0, p->pen.y);
		i = coordToIndex(p, 0, p->pen.y);
		vterm_clear(p, i, p->pen.x);
		break;
	case 2:
		vterm_text_clear_all(p);
		break;
	default:
		vterm_text_clean_band(p, p->pen.y + 1, p->ymax);
		i = coordToIndex(p, p->pen.x, p->pen.y);
		len = p->xcap - p->pen.x;
		vterm_clear(p, i, len);
		break;
	}
}

static void vterm_text_clean_band(TVterm *p, u_int top, u_int bottom)
{
	u_int y;
	u_int i;

	for (y = top; y < bottom; y++) {
		i = coordToIndex(p, 0, y);
		vterm_clear(p, i, p->xcap4); /* lclear */
	}
}

void vterm_text_scroll_down(TVterm *p, int line)
{
	vterm_text_move_down(p, p->ymin, p->ymax, line);
}

void vterm_text_move_down(TVterm *p, u_int top, u_int bottom, int line)
{
	u_int n;
	u_int src;
	u_int dst;

	if (bottom <= top + line)
		vterm_text_clean_band(p, top, bottom);
	else {
		for (n = bottom - 1; n >= top + line; n--) {
			dst = coordToIndex(p, 0, n);
			src = coordToIndex(p, 0, n - line);
			vterm_move(p, dst, src, p->xcap4); /* lmove */
			unclean4(p->flag + dst, p->xcap4);
		}
		vterm_text_clean_band(p, top, top + line);
	}
}

void vterm_text_scroll_up(TVterm *p, int line)
{
	vterm_text_move_up(p, p->ymin, p->ymax, line);
}

void vterm_text_move_up(TVterm *p, u_int top, u_int bottom, int line)
{
	u_int n;
	u_int src;
	u_int dst;

	if (bottom <= top + line)
		vterm_text_clean_band(p, top, bottom);
	else {
#ifdef ENABLE_SCROLLBACK
		for (n = top; n < top + line; n++) {
			src = coordToIndex(p, 0, n);
			scrollback_insertLine(p, src);
		}
#endif
		for (n = top; n < bottom - line; n++) {
			dst = coordToIndex(p, 0, n);
			src = coordToIndex(p, 0, n + line);
			vterm_move(p, dst, src, p->xcap4); /* lmove */
			unclean4(p->flag + dst, p->xcap4);
		}
		vterm_text_clean_band(p, bottom - line, bottom);
	}
}

#define SWAP(v1, v2, tmp) (tmp = v1, v1 = v2, v2 = tmp)

void vterm_reverseText(TVterm *p, u_int fx, u_int fy, u_int tx, u_int ty)
{
	u_int from, to, x, y, xx, tmp;
	u_char fcol, bcol, fcol2, bcol2;

	adjustCoord(p, &fx, &fy);
	adjustCoord(p, &tx, &ty);
	if (fy > ty) {
		SWAP(fy, ty, tmp);
		SWAP(fx, tx, tmp);
	} else if (fy == ty && fx > tx) {
		SWAP(fx, tx, tmp);
	}
	for (xx = p->xcap, y = fy; y <= ty; y++) {
		if (y == ty)
			xx = tx;
		from = coordToIndex(p, fx, y);
		to = coordToIndex(p, xx, y);
		if (p->flag[from] & VTERM_FLAG_CHAR_TAIL)
			from--;
		for (x = from; x <= to; x++) {
			if (p->rtxt[x] == 0)
				continue;
			fcol = p->fcol[x];
			bcol = p->bcol[x];
#ifdef ENABLE_256_COLOR
			if (fcol < 16 && bcol < 16) {
				fcol2 = (fcol & 8) | (bcol & ~8);
				bcol2 = (bcol & 8) | (fcol & ~8);
				p->fcol[x] = fcol2;
				p->bcol[x] = bcol2;
			} else {
				p->fcol[x] = bcol;
				p->bcol[x] = fcol;
			}
#else
			fcol2 = (fcol & 8) | (bcol & ~8);
			bcol2 = (bcol & 8) | (fcol & ~8);
			p->fcol[x] = fcol2;
			p->bcol[x] = bcol2;
#endif
			p->flag[x] &= ~VTERM_FLAG_CLEAN;
		}
		fx = 0;
	}
}

void vterm_copyText(TVterm *p, u_int fx, u_int fy, u_int tx, u_int ty)
{
	u_int from, to, x, y, xx, tmp;
	u_short rtxt;
	size_t nbytes;
	iconv_t mbcd, uccd;
	char inbytes[MAX_MULTIBYTE_LEN], outbytes[MAX_MULTIBYTE_LEN];
#if defined (__GLIBC__)
	char *inbuf;
#else
	const char *inbuf;
#endif
	char *outbuf;
	size_t inbytesleft, outbytesleft;

	clipboard_clear();

	mbcd = (iconv_t)-1;
	uccd = (iconv_t)-1;
#ifdef ENABLE_OTHER_CODING_SYSTEM
	if (vterm_is_otherCS(p)) {
		if (strcmp(p->otherCS->fromcode, p->otherCS->tocode) != 0)
			mbcd = iconv_open(p->otherCS->fromcode,
					  p->otherCS->tocode);
		uccd = iconv_open(p->otherCS->fromcode, "UCS-2BE");
	}
#endif
#ifdef ENABLE_UTF8
	if (uccd == (iconv_t)-1 && vterm_is_UTF8(p))
		uccd = iconv_open("UTF-8", "UCS-2BE");
#endif

	adjustCoord(p, &fx, &fy);
	adjustCoord(p, &tx, &ty);
	if (fy > ty) {
		SWAP(fy, ty, tmp);
		SWAP(fx, tx, tmp);
	} else if (fy == ty && fx > tx) {
		SWAP(fx, tx, tmp);
	}
	for (xx = p->xcap, y = fy; y <= ty; y++) {
		if (y == ty)
			xx = tx;
		from = coordToIndex(p, fx, y);
		if (p->flag[from] & VTERM_FLAG_CHAR_TAIL)
			from--;
		to = coordToIndex(p, xx, y);
		for (x = to; x > from; x--)
			if (p->rtxt[x] > 0)
				break;
		to = x;
		for (x = from; x <= to; x++) {
			rtxt = p->rtxt[x];
			outbytes[0] = 0x20;
			nbytes = 1;
			if (p->flag[x] & VTERM_FLAG_CHAR_SINGLE &&
			    p->flag[x] & VTERM_FLAG_MULTIBYTE) {
				outbytes[0] = rtxt & 0xff;
				nbytes = 1;
			} else if (p->flag[x] & VTERM_FLAG_CHAR_LEAD &&
				   p->flag[x] & VTERM_FLAG_MULTIBYTE) {
				if (mbcd != (iconv_t)-1) {
					inbytes[0] = (rtxt >> 8) & 0xff;
					inbytes[1] = rtxt & 0xff;
					inbuf = inbytes;
					inbytesleft = 2;
					outbuf = outbytes;
					outbytesleft = sizeof(outbytes);
					if (iconv(mbcd, &inbuf, &inbytesleft,
						  &outbuf, &outbytesleft) != -1)
						nbytes = sizeof(outbytes) -
							 outbytesleft;
					else {
						outbytes[0] = 0x20;
						nbytes = 1;
					}
				} else {
					outbytes[0] = (rtxt >> 8) & 0xff;
					outbytes[1] = rtxt & 0xff;
					nbytes = 2;
				}
			} else if ((p->flag[x] & VTERM_FLAG_CHAR_SINGLE &&
				    p->flag[x] & VTERM_FLAG_UNICODE) ||
				   (p->flag[x] & VTERM_FLAG_CHAR_LEAD &&
				    p->flag[x] & VTERM_FLAG_UNICODE)) {
				if (uccd != (iconv_t)-1) {
					inbytes[0] = (rtxt >> 8) & 0xff;
					inbytes[1] = rtxt & 0xff;
					inbuf = inbytes;
					inbytesleft = 2;
					outbuf = outbytes;
					outbytesleft = sizeof(outbytes);
					if (iconv(uccd, &inbuf, &inbytesleft,
						  &outbuf, &outbytesleft) != -1)
						nbytes = sizeof(outbytes) -
							 outbytesleft;
					else {
						outbytes[0] = 0x20;
						nbytes = 1;
					}
				}
			} else if (p->flag[x] & VTERM_FLAG_CHAR_TAIL)
				continue;
			if (nbytes == 1 &&
			    (outbytes[0] < 0x20 || outbytes[0] == 0x7f))
				outbytes[0] = 0x20;
			clipboard_appendText(outbytes, nbytes);
		}
		if (y < ty)
			clipboard_appendText("\n", 1);
		fx = 0;
	}
	if (mbcd != (iconv_t)-1)
		iconv_close(mbcd);
	if (uccd != (iconv_t)-1)
		iconv_close(uccd);
}

void vterm_pasteText(TVterm *p)
{
	if (clipboard_getLength() > 1)
		write_wrapper(p->term->masterPty,
		              clipboard_getText(), clipboard_getLength() - 1);
}

static void pollMouseCursor(TVterm *p)
{
	int x, y;
	bool wide;

	x = mouse_getX();
	y = mouse_getY();
	adjustCoord(p, &x, &y);
	wide = isLeadChar(p, x, y);
	if (mouseCursor.x != x || mouseCursor.y != y ||
	    mouseCursor.wide != wide) {
		cursor_show(p, &mouseCursor, false);
		mouseCursor.x = x;
		mouseCursor.y = y;
		mouseCursor.wide = wide;
		cursor_show(p, &mouseCursor, true);
	} else {
		if (mouse_isTimeout())
			cursor_show(p, &mouseCursor, false);
		else {
			if (mouseCursor.style == CURSOR_STYLE_BLINK &&
			    cursor_isTimeout(&mouseCursor)) {
				cursor_toggle(p, &mouseCursor);
				if (cursor.interval == mouseCursor.interval)
					cursor_resetTimer(&cursor);
			}
		}
	}
}

void vterm_pollCursor(TVterm *p, const bool wakeup)
{
	static bool scrollLocked[2];

	if (!p->active)
		return;
	if (wakeup) {
		screensaver_execute(false);
		cursor_show(p, &cursor, true);
		pollMouseCursor(p);
		return;
	}
	/* Idle. */
	if (screensaver_isRunning())
		return;
	if (screensaver_isEnable() && screensaver_isTimeout()) {
		cursor_show(p, &cursor, false);
		cursor_show(p, &mouseCursor, false);
		screensaver_execute(true);
		return;
	}
	if (cursor.style == CURSOR_STYLE_NORMAL) {
		scrollLocked[0] = keyboard_isScrollLocked();
		if (scrollLocked[0] != scrollLocked[1]) {
			cursor_show(p, &cursor, !scrollLocked[0]);
			scrollLocked[1] = scrollLocked[0];
			return;
		}
	} else if (cursor.style == CURSOR_STYLE_BLINK)
		if (cursor_isTimeout(&cursor))
			cursor_toggle(p, &cursor);
	if (mouse_isEnable())
		pollMouseCursor(p);
}

